from io import StringIO
from json import dumps as json_dumps
import pathlib
from subprocess import list2cmdline
from typing import Optional, Tuple

import click
import yaml

import anyscale
from anyscale.cli_logger import BlockLogger
from anyscale.commands.util import convert_kv_strings_to_dict
from anyscale.controllers.job_controller import JobController
from anyscale.job.models import JobConfig, JobState, JobStatus
from anyscale.util import validate_non_negative_arg


log = BlockLogger()  # CLI Logger


def _validate_job_name_and_id(name: Optional[str], job_id: Optional[str]):
    if name is None and job_id is None:
        raise click.ClickException("One of '--name' and '--job-id' must be provided.")

    if name is not None and job_id is not None:
        raise click.ClickException(
            "Only one of '--name' and '--job-id' can be provided."
        )


@click.group("job", help="Interact with production jobs running on Anyscale.")
def job_cli() -> None:
    pass


@job_cli.command(
    name="submit", short_help="Submit a job.",
)
@click.option("-n", "--name", required=False, default=None, help="Name of the job.")
@click.option(
    "-f",
    "--follow",
    required=False,
    hidden=True,
    default=False,
    type=bool,
    is_flag=True,
    help="DEPRECATED: use `--wait` instead.",
)
@click.option(
    "-w",
    "--wait",
    required=False,
    default=False,
    type=bool,
    is_flag=True,
    help="Block this CLI command and print logs until the job finishes.",
)
@click.option(
    "--config-file",
    required=False,
    default=None,
    type=str,
    help="Path to a YAML config file to use for this job. Command-line flags will overwrite values read from the file.",
)
@click.option(
    "--compute-config",
    required=False,
    default=None,
    type=str,
    help="Named compute configuration to use for the job. This defaults to the compute configuration of the workspace.",
)
@click.option(
    "--image-uri",
    required=False,
    default=None,
    type=str,
    help="Container image to use for this job. When running in a workspace, this defaults to the image of the workspace.",
)
@click.option(
    "--containerfile",
    required=False,
    default=None,
    type=str,
    help="Path to a containerfile to build the image to use for the job.",
)
@click.option(
    "--env",
    required=False,
    multiple=True,
    type=str,
    help="Environment variables to set for the job. The format is 'key=value'. This argument can be specified multiple times. When the same key is also specified in the config file, the value from the command-line flag will overwrite the value from the config file.",
)
@click.option(
    "--working-dir",
    required=False,
    default=None,
    type=str,
    help="Path to a local directory that will be the working directory for the job. The files in the directory will be automatically uploaded to cloud storage. When running in a workspace, this defaults to the current working directory.",
)
@click.option(
    "-e",
    "--exclude",
    required=False,
    type=str,
    multiple=True,
    help="File pattern to exclude when uploading local directories. This argument can be specified multiple times and the patterns will be appended to the 'excludes' list in the config file (if any).",
)
@click.option(
    "-r",
    "--requirements",
    required=False,
    default=None,
    type=str,
    help="Path to a requirements.txt file containing dependencies for the job. These will be installed on top of the image. When running in a workspace, this defaults to the workspace dependencies.",
)
@click.option(
    "--py-module",
    required=False,
    default=None,
    multiple=True,
    type=str,
    help="Python modules to be available for import in the Ray workers. Each entry must be a path to a local directory.",
)
@click.argument("entrypoint", required=False, nargs=-1, type=click.UNPROCESSED)
def submit(  # noqa: PLR0912 PLR0913 C901
    entrypoint: Tuple[str],
    name: Optional[str],
    follow: Optional[bool],
    wait: Optional[bool],
    config_file: Optional[str],
    compute_config: Optional[str],
    image_uri: Optional[str],
    containerfile: Optional[str],
    env: Tuple[str],
    working_dir: Optional[str],
    exclude: Tuple[str],
    requirements: Optional[str],
    py_module: Tuple[str],
):
    """Submit a job.

    The job config can be specified in one of the following ways:

    * Job config file can be specified as a single positional argument. E.g. `anyscale job submit config.yaml`.

    * Job config can also be specified with command-line arguments. In this case, the entrypoint should be specified
as the positional arguments starting with `--`. Other arguments can be specified with command-line flags. E.g.

      * `anyscale job submit -- python main.py`: submit a job with the entrypoint `python main.py`.

      * `anyscale job submit --name my-job -- python main.py`: submit a job with the name `my-job` and the
entrypoint `python main.py`.

    * [Experimental] If you want to specify a config file and override some arguments with the commmand-line flags,
use the `--config-file` flag. E.g.

      * `anyscale job submit --config-file config.yaml`: submit a job with the config in `config.yaml`.

      * `anyscale job submit --config-file config.yaml -- python main.py`: submit a job with the config in `config.yaml`
and override the entrypoint with `python main.py`.

    Either containerfile or image-uri should be used, specifying both will result in an error.

    By default, this command submits the job asynchronously and exits. To wait for the job to complete, use the `--wait` flag.
    """

    job_controller = JobController()
    if len(entrypoint) == 1 and (
        pathlib.Path(entrypoint[0]).is_file() or entrypoint[0].endswith(".yaml")
    ):
        # If entrypoint is a single string that ends with .yaml, e.g. `anyscale job submit config.yaml`,
        # treat it as a config file, and use the old job submission API.
        if config_file is not None:
            raise click.ClickException(
                "`--config-file` should not be used when providing a config file as the entrypoint."
            )
        if image_uri:
            raise click.ClickException(
                "`--image-uri` should not be used when providing a config file as the entrypoint."
            )
        if containerfile:
            raise click.ClickException(
                "`--containerfile` should not be used when providing a config file as the entrypoint."
            )
        if env:
            raise click.ClickException(
                "`--env` should not be used when providing a config file as the entrypoint."
            )

        config_file = entrypoint[0]
        if not pathlib.Path(config_file).is_file():
            raise click.ClickException(f"Job config file '{config_file}' not found.")
        log.info(f"Submitting job from config file {config_file}.")

        job_id = job_controller.submit(config_file, name=name)
    else:
        # Otherwise, use the new job submission API. E.g.
        # `anyscale job submit -- python main.py`,
        # or `anyscale job submit --config-file config.yaml`.
        if len(entrypoint) == 0 and config_file is None:
            raise click.ClickException(
                "Either a config file or an inlined entrypoint must be provided."
            )
        if config_file is not None and not pathlib.Path(config_file).is_file():
            raise click.ClickException(f"Job config file '{config_file}' not found.")

        if follow:
            raise click.ClickException(
                "`--follow` is deprecated, use `--wait` instead."
            )

        args = {}
        if len(entrypoint) > 0:
            args["entrypoint"] = list2cmdline(entrypoint)
        if name:
            args["name"] = name

        if containerfile and image_uri:
            raise click.ClickException(
                "Only one of '--containerfile' and '--image-uri' can be provided."
            )

        if image_uri:
            args["image_uri"] = image_uri

        if containerfile:
            args["containerfile"] = containerfile

        if working_dir:
            args["working_dir"] = working_dir

        if config_file is not None:
            config = JobConfig.from_yaml(config_file, **args)
        else:
            config = JobConfig.from_dict(args)

        if compute_config is not None:
            config = config.options(compute_config=compute_config)

        if exclude:
            config = config.options(excludes=[e for e in exclude])

        if requirements is not None:
            if not pathlib.Path(requirements).is_file():
                raise click.ClickException(
                    f"Requirements file '{requirements}' not found."
                )
            config = config.options(requirements=requirements)

        if env:
            env_dict = convert_kv_strings_to_dict(env)
            if env_dict:
                config = config.options(env_vars=env_dict)
        if py_module:
            for module in py_module:
                if not pathlib.Path(module).is_dir():
                    raise click.ClickException(
                        f"Python module path '{module}' does not exist or is not a directory."
                    )
            config = config.options(py_modules=[*py_module])
        log.info(f"Submitting job with config {config}.")
        job_id = anyscale.job.submit(config)

    if wait:
        log.info(
            "Waiting for the job to run. Interrupting this command will not cancel the job."
        )
    else:
        log.info("Use `--wait` to wait for the job to run and stream logs.")

    if wait:
        anyscale.job.wait(job_id=job_id)


@job_cli.command(name="list", help="Display information about existing jobs.")
@click.option("--name", "-n", required=False, default=None, help="Filter by job name.")
@click.option(
    "--job-id", "--id", required=False, default=None, help="Filter by job id."
)
@click.option(
    "--project-id", required=False, default=None, help="Filter by project id."
)
@click.option(
    "--include-all-users",
    is_flag=True,
    default=False,
    help="Include jobs not created by current user.",
)
@click.option(
    "--include-archived",
    is_flag=True,
    default=False,
    help=(
        "List archived jobs as well as unarchived jobs."
        "If not provided, defaults to listing only unarchived jobs."
    ),
)
@click.option(
    "--max-items",
    required=False,
    default=10,
    type=int,
    help="Max items to show in list.",
    callback=validate_non_negative_arg,
)
def list(  # noqa: A001
    name: Optional[str],
    job_id: Optional[str],
    project_id: Optional[str],
    include_all_users: bool,
    include_archived: bool,
    max_items: int,
) -> None:
    job_controller = JobController()
    job_controller.list(
        name=name,
        job_id=job_id,
        project_id=project_id,
        include_all_users=include_all_users,
        include_archived=include_archived,
        max_items=max_items,
    )


@job_cli.command(name="archive", short_help="Archive a job.")
@click.option("--job-id", "--id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
def archive(job_id: Optional[str], name: Optional[str]) -> None:
    """Archive a job.

    To specify the job by name, use the --name flag. To specify the job by id, use the --job-id flag. Either name or
id should be used, specifying both will result in an error.

    If job is specified by name and there are multiple jobs with the specified name, the most recently created job
status will be archived.
    """
    _validate_job_name_and_id(name=name, job_id=job_id)
    anyscale.job.archive(job_id=job_id, name=name)


@job_cli.command(name="terminate", short_help="Terminate a job.")
@click.option("--job-id", "--id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
def terminate(job_id: Optional[str], name: Optional[str]) -> None:
    """Terminate a job.

    To specify the job by name, use the --name flag. To specify the job by id, use the --job-id flag. Either name or
id should be used, specifying both will result in an error.

    If job is specified by name and there are multiple jobs with the specified name, the most recently created job
status will be terminated.
    """
    _validate_job_name_and_id(name=name, job_id=job_id)
    anyscale.job.terminate(name=name, job_id=job_id)
    if job_id is not None:
        log.info(
            f"Query the status of the job with `anyscale job status --id {job_id}`."
        )
    else:
        log.info(
            f"Query the status of the job with `anyscale job status --name {name}`."
        )


@job_cli.command(name="logs")
@click.option("--job-id", "--id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
@click.option(
    "--follow",
    "-f",
    required=False,
    default=False,
    type=bool,
    is_flag=True,
    help="Whether to follow the log.",
)
@click.option(
    "--all-attempts",
    is_flag=True,
    default=False,
    help="Show logs for all job attempts.",
)
def logs(
    job_id: Optional[str],
    name: Optional[str],
    follow: bool = False,
    all_attempts: bool = False,
) -> None:
    """Print the logs of a job.

    By default from the latest job attempt.

    Example usage:

        anyscale job logs --id prodjob_123

        anyscale job logs --id prodjob_123 -f

        anyscale job logs -n my-job --all-attempts"""
    job_controller = JobController(raise_structured_exception=True)
    job_controller.logs(
        job_id=job_id, job_name=name, should_follow=follow, all_attempts=all_attempts,
    )


@job_cli.command(name="wait", short_help="Wait for a job to enter a specific state.")
@click.option("--job-id", "--id", required=False, help="Unique ID of the job.")
@click.option("--name", "-n", required=False, help="Name of the job.")
@click.option(
    "--state",
    "-s",
    required=False,
    default=JobState.SUCCEEDED,
    help="The state to wait for this job to enter",
)
@click.option(
    "--timeout-s",
    "--timeout",
    "-t",
    required=False,
    default=1800,
    type=float,
    help="The timeout in seconds after which this command will exit.",
)
def wait(
    job_id: Optional[str],
    name: Optional[str],
    state: str = JobState.SUCCEEDED,
    timeout_s=None,
) -> None:
    """Wait for a job to enter a specific state (default: SUCCEEDED).

    To specify the job by name, use the --name flag. To specify the job by id, use the --job-id flag.

    If the job reaches the target state, the command will exit successfully.

    If the job reaches a terminal state other than the target state, the command will exit with an error.

    If the command reaches the timeout, the command will exit with an error but job execution will continue.
    """
    try:
        state = JobState.validate(state)
    except ValueError as e:
        raise click.ClickException(str(e))
    anyscale.job.wait(name=name, job_id=job_id, state=state, timeout_s=timeout_s)


@job_cli.command(
    name="status", short_help="Get the status of a job.",
)
@click.option(
    "--job-id", "--id", required=False, default=None, help="Unique ID of the job."
)
@click.option("--name", "-n", required=False, default=None, help="Name of the job.")
@click.option(
    "--json",
    "-j",
    is_flag=True,
    default=False,
    help="Output the status in a structured JSON format.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Include verbose details in the status.",
)
def status(name: Optional[str], job_id: Optional[str], json: bool, verbose: bool):
    """Query the status of a job.

    To specify the job by name, use the --name flag. To specify the job by id, use the --job-id flag. Either name or
id should be used, specifying both will result in an error.

    If job is specified by name and there are multiple jobs with the specified name, the most recently created job
status will be returned.
    """
    _validate_job_name_and_id(name=name, job_id=job_id)

    status: JobStatus = anyscale.job.status(name=name, job_id=job_id)
    status_dict = status.to_dict()

    if not verbose:
        status_dict.pop("config", None)

    if json:
        print(json_dumps(status_dict, indent=4, sort_keys=False))
    else:
        stream = StringIO()
        yaml.dump(status_dict, stream, sort_keys=False)
        print(stream.getvalue(), end="")



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
import asyncio
from asyncio import Future, isfuture
import logging
from typing import Awaitable, List, Optional, TypeVar, Union


_T = TypeVar("_T")
logger = logging.getLogger(__file__)


def create_future() -> Future:
    """Creates instance of `asyncio.Future` using the currently running event-loop"""
    loop = asyncio.get_running_loop()
    return loop.create_future()


def chain_future(source, destination):
    """NOTE: THIS METHOD IS COPIED FROM STDLIB VERBATIM, WITH NO MODIFICATIONS, PLEASE KEEP IT THAT WAY.
             THIS WAS COPIED BECAUSE ITS PRIVATE IN STANDARD LIBRARY MAKING IT IMPOSSIBLE TO USE IT
             OUTSIDE OF IT.

    Chain two futures so that when one completes, so does the other.

    The result (or exception) of source will be copied to destination.
    If destination is cancelled, source gets cancelled too.
    Compatible only with `asyncio.Future`
    """
    if not isfuture(source):
        raise TypeError("A future is required for source argument")
    if not isfuture(destination):
        raise TypeError("A future is required for destination argument")
    source_loop = _get_loop(source) if isfuture(source) else None
    dest_loop = _get_loop(destination) if isfuture(destination) else None

    def _set_state(future, other):
        _copy_future_state(other, future)

    def _call_check_cancel(destination):
        if destination.cancelled():
            if source_loop is None or source_loop is dest_loop:
                source.cancel()
            else:
                source_loop.call_soon_threadsafe(source.cancel)

    def _call_set_state(source):
        if destination.cancelled() and dest_loop is not None and dest_loop.is_closed():
            return
        if dest_loop is None or dest_loop is source_loop:
            _set_state(destination, source)
        else:
            dest_loop.call_soon_threadsafe(_set_state, destination, source)

    destination.add_done_callback(_call_check_cancel)
    source.add_done_callback(_call_set_state)


def _get_loop(fut):
    """NOTE: THIS METHOD IS COPIED FROM STDLIB VERBATIM, WITH NO MODIFICATIONS, PLEASE KEEP IT THAT WAY.
             THIS WAS COPIED BECAUSE ITS PRIVATE IN STANDARD LIBRARY MAKING IT IMPOSSIBLE TO USE IT
             OUTSIDE OF IT.
    """
    # Tries to call Future.get_loop() if it's available.
    # Otherwise fallbacks to using the old '_loop' property.
    try:
        get_loop = fut.get_loop
    except AttributeError:
        pass
    else:
        return get_loop()
    return fut._loop  # noqa: SLF001


def _copy_future_state(source, dest):
    """NOTE: THIS METHOD IS COPIED FROM STDLIB VERBATIM, WITH NO MODIFICATIONS, PLEASE KEEP IT THAT WAY.
             THIS WAS COPIED BECAUSE ITS PRIVATE IN STANDARD LIBRARY MAKING IT IMPOSSIBLE TO USE IT
             OUTSIDE OF IT.

    Internal helper to copy state from another Future.

    The other Future may be a concurrent.futures.Future.
    """
    assert source.done()
    if dest.cancelled():
        return
    assert not dest.done()
    if source.cancelled():
        dest.cancel()
    else:
        exception = source.exception()
        if exception is not None:
            dest.set_exception(exception)
        else:
            result = source.result()
            dest.set_result(result)


async def gather_in_batches(
    n: int,
    *tasks: Awaitable[_T],
    per_task_timeout_secs: Optional[float] = None,
    return_exceptions: bool = False,
    log_exceptions: bool = False,
) -> List[Union[_T, BaseException]]:
    # n cannot be less than 1
    n = max(n, 1)
    semaphore = asyncio.Semaphore(n)

    async def sem_task(task: Awaitable[_T]) -> _T:
        try:
            async with semaphore:
                return await asyncio.wait_for(task, per_task_timeout_secs)
        except Exception:
            if log_exceptions:
                logger.exception("Error while gathering task")
            raise  # re-raise exception

    return await asyncio.gather(
        *(sem_task(task) for task in tasks), return_exceptions=return_exceptions
    )



from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple

from anyscale._private.models.image_uri import ImageURI
from anyscale.client.openapi_client.models import (
    Cloud,
    ComputeTemplateConfig,
    CreateInternalProductionJob,
    DecoratedComputeTemplate,
    InternalProductionJob,
)
from anyscale.client.openapi_client.models.production_job import ProductionJob
from anyscale.sdk.anyscale_client.models import (
    ApplyServiceModel,
    Cluster,
    ClusterCompute,
    ServiceModel,
)
from anyscale.utils.workspace_notification import WorkspaceNotification


# TODO(edoakes): figure out a sane policy for this.
# Maybe just make it part of the release process to update it, or fetch the
# default builds and get the latest one. The best thing to do is probably
# to populate this in the backend.
DEFAULT_RAY_VERSION = "2.10.0"
DEFAULT_PYTHON_VERSION = "py310"
RUNTIME_ENV_PACKAGE_FORMAT = "pkg_{content_hash}.zip"

# All workspace cluster names should start with this prefix.
WORKSPACE_CLUSTER_NAME_PREFIX = "workspace-cluster-"


class AnyscaleClientInterface(ABC):
    @abstractmethod
    def get_job_ui_url(self, job_id: str) -> str:
        """Get a URL to the webpage for a job."""
        raise NotImplementedError

    @abstractmethod
    def get_service_ui_url(self, service_id: str) -> str:
        """Get a URL to the webpage for a service."""
        raise NotImplementedError

    @abstractmethod
    def get_current_workspace_id(self) -> Optional[str]:
        """Returns the ID of the workspace this is running in (or `None`)."""
        raise NotImplementedError

    @abstractmethod
    def inside_workspace(self) -> bool:
        """Returns true if this code is running inside a workspace."""
        raise NotImplementedError

    @abstractmethod
    def get_workspace_requirements_path(self) -> Optional[str]:
        """Returns the path to the workspace-managed requirements file.

        Returns None if dependency tracking is disable or the file does not exist or is not in a workspace.
        """
        raise NotImplementedError

    @abstractmethod
    def get_workspace_env_vars(self) -> Optional[Dict[str, str]]:
        """Returns the environment variables specified in workspace runtime dependencies."""
        raise NotImplementedError

    @abstractmethod
    def get_current_workspace_cluster(self) -> Optional[Cluster]:
        """Get the cluster model for the workspace this code is running in.

        Returns None if not running in a workspace.
        """
        raise NotImplementedError

    def send_workspace_notification(self, notification: WorkspaceNotification):
        """Send a workspace notification to be displayed to the user.

        This is a no-op if called from outside a workspace.
        """
        raise NotImplementedError

    @abstractmethod
    def get_project_id(self, parent_cloud_id: str) -> str:
        """Get the default project ID for this user.

        If running in a workspace, returns the workspace project ID.
        """
        raise NotImplementedError

    @abstractmethod
    def get_cloud_id(
        self, *, cloud_name: Optional[str] = None, compute_config_id: Optional[str]
    ) -> str:
        """Get the cloud ID for the provided cloud name or compute config ID.

        If both arguments are None:
            - if running in a workspace, get the workspace's cloud ID.
            - else, get the user's default cloud ID.
        """
        raise NotImplementedError

    @abstractmethod
    def get_cloud(self, *, cloud_id: str) -> Optional[Cloud]:
        """Get the cloud model for the provided cloud ID.

        Returns `None` if the cloud ID was not found.
        """
        raise NotImplementedError

    @abstractmethod
    def create_compute_config(
        self, config: ComputeTemplateConfig, *, name: Optional[str] = None
    ) -> Tuple[str, str]:
        """Create a compute config and return its ID.

        If a name is not provided, the compute config will be "anonymous."

        Returns (name, ID).
        """
        raise NotImplementedError

    @abstractmethod
    def get_compute_config_id(
        self, compute_config_name: Optional[str] = None, *, include_archived=False
    ) -> Optional[str]:
        """Get the compute config ID for the provided name.

        If compute_config_name is None:
            - if running in a workspace, get the workspace's compute config ID.
            - else, get the user's default compute config ID.

        Returns None if the compute config name does not exist.
        """
        raise NotImplementedError

    @abstractmethod
    def get_compute_config(
        self, compute_config_id: str
    ) -> Optional[DecoratedComputeTemplate]:
        """Get the compute config for the provided ID.

        Returns None if not found.
        """
        raise NotImplementedError

    @abstractmethod
    def archive_compute_config(self, *, compute_config_id: str):
        """Archive the compute config for the provided ID."""
        raise NotImplementedError

    @abstractmethod
    def get_default_compute_config(self, *, cloud_id: str) -> ClusterCompute:
        """Get the default compute config for the provided cloud ID."""
        raise NotImplementedError

    @abstractmethod
    def get_default_build_id(self) -> str:
        """Get the default build id.

        If running in a workspace, it returns the workspace build id.
        Else it returns the default build id.
        """
        raise NotImplementedError

    @abstractmethod
    def build_image_from_requirements(
        self, cluster_env_name: str, base_build_id: str, requirements: List[str]
    ) -> str:
        """Build an image with the provided requirements and base_build_id.
        """
        raise NotImplementedError

    @abstractmethod
    def get_cluster_env_build_id_from_containerfile(
        self, cluster_env_name: str, containerfile: str
    ) -> str:
        """Get the cluster environment build ID for the cluster environment with provided containerfile.
        Look for an existing cluster environment with the provided name.
        If found, reuse it. Else, create a new cluster environment.
        Create a new cluster environment build with the provided containerfile or try to reuse one with the same containerfile if exists.
        """
        raise NotImplementedError

    @abstractmethod
    def get_cluster_env_build_id_from_image_uri(
        self, image_uri: ImageURI,
    ):
        """Get the cluster environment build ID for the cluster environment with provided image_uri.

        It maps a image_uri to a cluster env name and then use the name to get the cluster env.
        If there exists a cluster environment for the image_uri, it will reuse the cluster env.
        Else it will create a new cluster environment.
        It will create a new cluster environment build with the provided image_uri or try to reuse one with the same image_uri if exists.

        The same URI should map to the same cluster env name and therefore the build but it is not guaranteed since
        the name format can change.

        """
        raise NotImplementedError

    @abstractmethod
    def get_cluster_env_build_image_uri(
        self, cluster_env_build_id: str
    ) -> Optional[ImageURI]:
        """Get the image URI for the provided build ID.

        Returns None if not found.
        """
        raise NotImplementedError

    @abstractmethod
    def get_service(self, name: str) -> Optional[ServiceModel]:
        """Get a service by name.

        Returns None if not found.
        """
        raise NotImplementedError

    @abstractmethod
    def get_job(
        self, name: Optional[str], job_id: Optional[str]
    ) -> Optional[ProductionJob]:
        """Get a job by name.

        Returns None if not found.
        """
        raise NotImplementedError

    @abstractmethod
    def rollout_service(self, model: ApplyServiceModel) -> ServiceModel:
        """Deploy or update the service to use the provided config.

        Returns the service ID.
        """
        raise NotImplementedError

    @abstractmethod
    def rollback_service(
        self, service_id: str, *, max_surge_percent: Optional[int] = None
    ):
        """Roll the service back to the primary version.

        This can only be used during an active rollout.
        """
        raise NotImplementedError

    @abstractmethod
    def terminate_service(self, service_id: str):
        """Mark the service to be terminated asynchronously."""
        raise NotImplementedError

    @abstractmethod
    def submit_job(self, model: CreateInternalProductionJob) -> InternalProductionJob:
        """Submit the job to run."""
        raise NotImplementedError

    @abstractmethod
    def terminate_job(self, job_id: str):
        """Mark the job to be terminated asynchronously."""
        raise NotImplementedError

    @abstractmethod
    def archive_job(self, job_id: str):
        """Mark the job to be archived asynchronously."""
        raise NotImplementedError

    @abstractmethod
    def upload_local_dir_to_cloud_storage(
        self,
        local_dir: str,
        *,
        cloud_id: str,
        excludes: Optional[List[str]] = None,
        overwrite_existing_file: bool = False
    ) -> str:
        """Upload the provided directory to cloud storage and return a URI for it.

        The directory will be zipped and the resulting URI can be used in a Ray runtime_env.

        The upload is preformed using a pre-signed URL fetched from Anyscale, so no
        local cloud provider authentication is required.

        The URI is content-addressable (containing a hash of the directory contents), so by
        default if the target file URI already exists it will not be overwritten.
        """
        raise NotImplementedError



from dataclasses import dataclass, field
from typing import Optional, Union

from anyscale._private.models import ModelBase, ModelEnum
from anyscale._private.workload import WorkloadConfig


@dataclass(frozen=True)
class JobConfig(WorkloadConfig):
    """Configuration options for a job."""

    __doc_py_example__ = """\
from anyscale.job.models import JobConfig

config = JobConfig(
    name="my-job",
    entrypoint="python main.py",
    max_retries=1,
    # An inline `ComputeConfig` can also be provided.
    compute_config="my-compute-config:1",
    # A containerfile path can also be provided.
    image_uri="anyscale/image/my-image:1",
)
"""

    __doc_yaml_example__ = """\
name: my-job
entrypoint: python main.py
# An inline dictionary can also be provided.
compute_config: my-compute-config:1
# A containerfile path can also be provided.
image_uri: anyscale/image/my-image:1
"""

    # Override the `name` field from `WorkloadConfig` so we can document it separately for jobs and services.
    name: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "Name of the job. Multiple jobs can be submitted with the same name."
        },
    )

    entrypoint: str = field(
        default="",
        repr=False,
        metadata={
            "docstring": "Command that will be run to execute the job, e.g., `python main.py`."
        },
    )

    def _validate_entrypoint(self, entrypoint: str):
        if not isinstance(entrypoint, str):
            raise TypeError("'entrypoint' must be a string.")

        if not entrypoint:
            raise ValueError("'entrypoint' cannot be empty.")

    max_retries: int = field(
        default=1,
        repr=False,
        metadata={
            "docstring": "Maximum number of times the job will be retried before being marked failed. Defaults to `1`."
        },
    )

    def _validate_max_retries(self, max_retries: int):
        if not isinstance(max_retries, int):
            raise TypeError("'max_retries' must be an int.")

        if max_retries < 0:
            raise ValueError("'max_retries' must be >= 0.")


class JobState(ModelEnum):
    """Current state of a job."""

    STARTING = "STARTING"
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    SUCCEEDED = "SUCCEEDED"
    UNKNOWN = "UNKNOWN"

    _TERMINAL_JOB_STATES = [
        SUCCEEDED,
        FAILED,
    ]

    def __str__(self):
        return self.name

    @classmethod
    def is_terminal(cls, state: "JobState"):
        return state in cls._TERMINAL_JOB_STATES

    __docstrings__ = {
        STARTING: "The job is being started and is not yet running.",
        RUNNING: "The job is running. A job will have state RUNNING if a job run fails and there are remaining retries.",
        FAILED: "The job did not finish running or the entrypoint returned an exit code other than 0 after retrying up to max_retries times.",
        SUCCEEDED: "The job finished running and its entrypoint returned exit code 0.",
        UNKNOWN: "The CLI/SDK received an unexpected state from the API server. In most cases, this means you need to update the CLI.",
    }


@dataclass(frozen=True)
class JobStatus(ModelBase):
    """Current status of a job."""

    __doc_py_example__ = """\
import anyscale
from anyscale.job.models import JobStatus
status: JobStatus = anyscale.job.status(name="my-job")
"""

    __doc_cli_example__ = """\
$ anyscale job status -n my-job
id: prodjob_3suiybn8r7dhz92yv63jqzm473
name: my-job
state: STARTING
"""

    id: str = field(
        metadata={
            "docstring": "Unique ID of the job (generated when the job is first submitted)."
        }
    )

    def _validate_id(self, id: str):  # noqa: A002
        if not isinstance(id, str):
            raise TypeError("'id' must be a string.")

    name: str = field(
        metadata={
            "docstring": "Name of the job. Multiple jobs can be submitted with the same name."
        },
    )

    def _validate_name(self, name: str):
        if not isinstance(name, str):
            raise TypeError("'name' must be a string.")

    state: Union[str, JobState] = field(
        metadata={"docstring": "Current state of the job."}
    )

    def _validate_state(self, state: Union[str, JobState]) -> JobState:
        if isinstance(state, str):
            # This will raise a ValueError if the state is unrecognized.
            state = JobState(state)
        elif not isinstance(state, JobState):
            raise TypeError("'state' must be a JobState.")

        return state

    config: JobConfig = field(
        repr=False, metadata={"docstring": "Configuration of the job."}
    )

    def _validate_config(self, config: JobConfig):
        if not isinstance(config, JobConfig):
            raise TypeError("'config' must be a JobConfig.")



import json

from anyscale.aws_iam_policies import ANYSCALE_IAM_POLICIES


def test_policy_sid_unique():
    """
    Please do not change this test.

    This test is to ensure
    1) The policy version is 2012-10-17
    2) The policy statement has a Sid
    3) The Sid is unique
    """
    policy_sids = []
    for policy in ANYSCALE_IAM_POLICIES:
        policy_document = json.loads(policy.policy_document)
        assert policy_document["Version"] == "2012-10-17"
        for statement in policy_document["Statement"]:
            assert "Sid" in statement
            policy_sids.append(statement["Sid"])
    assert len(policy_sids) == len(set(policy_sids))



from typing import Dict, Tuple

import click


# Take from https://stackoverflow.com/questions/51846634/click-dynamic-defaults-for-prompts-based-on-other-options
class OptionPromptNull(click.Option):
    """
    Option class that allows default values based on previous params
    """

    _value_key = "_default_val"

    def __init__(self, *args, **kwargs):
        self.default_option = kwargs.pop("default_option", None)
        super().__init__(*args, **kwargs)

    def get_default(self, ctx, **kwargs):
        if not hasattr(self, self._value_key):
            if self.default_option is None:
                default = super().get_default(ctx, **kwargs)
            else:
                arg = ctx.params.get(self.default_option)
                if arg is not None:
                    default = self.type_cast_value(ctx, self.default(arg))
                else:
                    default = None
            setattr(self, self._value_key, default)
        return getattr(self, self._value_key)


# Taken from https://stackoverflow.com/questions/44247099/click-command-line-interfaces-make-options-required-if-other-optional-option-is
class NotRequiredIf(click.Option):
    """
    Option class that allows an option to be not required if a separate option is available.
    """

    def __init__(self, *args, **kwargs):
        self.not_required_if = kwargs.pop("not_required_if")
        assert self.not_required_if, "'not_required_if' parameter required"
        kwargs["help"] = (
            kwargs.get("help", "")
            + " NOTE: This argument is mutually exclusive with %s"
            % self.not_required_if
        ).strip()
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        we_are_present = self.name in opts
        other_present = self.not_required_if in opts

        if other_present:
            if we_are_present:
                raise click.UsageError(
                    "Illegal usage: `%s` is mutually exclusive with `%s`"
                    % (self.name, self.not_required_if)
                )
            else:
                self.prompt = None
                self.default = None

        return super().handle_parse_result(ctx, opts, args)


def convert_kv_strings_to_dict(strings: Tuple[str]) -> Dict[str, str]:
    """Convert args/env_vars of the form "key=val" into a dictionary of {key: val}.

    NOTE(edoakes): this mimics the functionality of the `serve run` CLI and should be
    kept in sync with it.
    """
    ret_dict = {}
    for s in strings:
        split = s.split("=")
        if len(split) != 2:
            raise click.ClickException(
                f"Invalid key-value string '{s}'. Must be of the form 'key=value'."
            )

        ret_dict[split[0]] = split[1]

    return ret_dict



from anyscale._private.sdk import sdk_command
from anyscale.compute_config._private.compute_config_sdk import ComputeConfigSDK
from anyscale.compute_config.models import ComputeConfig


_COMPUTE_CONFIG_SDK_SINGLETON_KEY = "compute_config_sdk"

_CREATE_EXAMPLE = """
import anyscale
from anyscale.compute_config.models import ComputeConfig, HeadNodeConfig, MarketType, WorkerNodeGroupConfig

config = ComputeConfig(
    head_node=HeadNodeConfig(
        instance_type="m5.8xlarge",
    ),
    worker_nodes=[
        WorkerNodeGroupConfig(
            instance_type="m5.8xlarge",
            min_nodes=5,
            max_nodes=5,
        ),
        WorkerNodeGroupConfig(
            instance_type="m5.4xlarge",
            min_nodes=1,
            max_nodes=10,
            market_type=MarketType.SPOT,
        ),
    ],
)
full_name: str = anyscale.compute_config.create(config, name="my-compute-config")
"""


@sdk_command(
    _COMPUTE_CONFIG_SDK_SINGLETON_KEY,
    ComputeConfigSDK,
    doc_py_example=_CREATE_EXAMPLE,
    arg_docstrings={
        "config": "The config options defining the compute config.",
        "name": "The name of the compute config. This should *not* include a version tag.",
    },
)
def create(config: ComputeConfig, *, name: str, _sdk: ComputeConfigSDK,) -> str:
    """Create a compute config under the specified name.

    Returns the full name of the registered compute config, including the version.
    """
    full_name, _ = _sdk.create_compute_config(config, name=name,)
    return full_name


_GET_EXAMPLE = """
import anyscale
from anyscale.compute_config.models import ComputeConfig

compute_config: ComputeConfig = anyscale.compute_config.get("my-compute-config")
"""


@sdk_command(
    _COMPUTE_CONFIG_SDK_SINGLETON_KEY,
    ComputeConfigSDK,
    doc_py_example=_GET_EXAMPLE,
    arg_docstrings={
        "name": "The name of the compute config. This can inclue an optional version tag (`name:version`). If no version tag is provided, the latest version will be returned.",
        "include_archived": "Whether to consider archived compute configs (defaults to False).",
    },
)
def get(
    name: str, *, include_archived: bool = False, _sdk: ComputeConfigSDK,
) -> ComputeConfig:
    """Get the compute config with the specified name.

    The name can contain an optional version, e.g., 'name:version'.
    If no version is provided, the latest one will be returned.
    """
    return _sdk.get_compute_config(name=name, include_archived=include_archived)


_ARCHIVE_EXAMPLE = """
import anyscale

anyscale.compute_config.archive(name="my-compute-config")
"""


@sdk_command(
    _COMPUTE_CONFIG_SDK_SINGLETON_KEY,
    ComputeConfigSDK,
    doc_py_example=_ARCHIVE_EXAMPLE,
    arg_docstrings={"name": "Name of the compute config."},
)
def archive(name: str = None, *, _sdk: ComputeConfigSDK) -> str:
    """Archive a compute config and all of its versions.

    Once a compute config is archived, its name will no longer be usable in the organization.
    """
    return _sdk.archive_compute_config(name=name)



import copy
import os
import pathlib
from typing import Any, Dict, List, Optional, Tuple, Union

from anyscale._private.anyscale_client import (
    AnyscaleClientInterface,
    WORKSPACE_CLUSTER_NAME_PREFIX,
)
from anyscale._private.sdk.base_sdk import BaseSDK, Timer
from anyscale.cli_logger import BlockLogger
from anyscale.compute_config._private.compute_config_sdk import ComputeConfigSDK
from anyscale.compute_config.models import ComputeConfig
from anyscale.utils.runtime_env import is_dir_remote_uri, parse_requirements_file


ENABLE_IMAGE_BUILD_FOR_TRACKED_REQUIREMENTS = (
    os.environ.get("ANYSCALE_ENABLE_IMAGE_BUILD_FOR_TRACKED_REQUIREMENTS", "0") == "1"
)


class WorkloadSDK(BaseSDK):
    """Shared parent class for job and service SDKs."""

    def __init__(
        self,
        *,
        logger: Optional[BlockLogger] = None,
        client: Optional[AnyscaleClientInterface] = None,
        timer: Optional[Timer] = None,
    ):
        super().__init__(logger=logger, client=client, timer=timer)
        self._enable_image_build_for_tracked_requirements = (
            ENABLE_IMAGE_BUILD_FOR_TRACKED_REQUIREMENTS
        )
        self._compute_config_sdk = ComputeConfigSDK(
            logger=self.logger, client=self.client,
        )

    @property
    def enable_image_build_for_tracked_requirements(self) -> bool:
        return self._enable_image_build_for_tracked_requirements

    def update_env_vars(
        self,
        runtime_envs: List[Dict[str, Any]],
        *,
        env_vars_updates: Optional[Dict[str, str]],
    ) -> List[Dict[str, Any]]:
        """Replaces 'env_vars' fields in runtime_envs with the override."""
        new_runtime_envs = copy.deepcopy(runtime_envs)

        if env_vars_updates:
            for runtime_env in new_runtime_envs:
                if "env_vars" in runtime_env:
                    # the precedence should be config > runtime_env
                    runtime_env["env_vars"].update(env_vars_updates)
                else:
                    runtime_env["env_vars"] = env_vars_updates

        return new_runtime_envs

    def override_and_load_requirements_files(
        self,
        runtime_envs: List[Dict[str, Any]],
        *,
        requirements_override: Union[None, str, List[str]],
        workspace_requirements_path: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """Replaces 'pip' fields in runtime_envs with their parsed file contents.

        The precedence for overrides is: explicit overrides passed in > fields in the existing
        runtime_envs > workspace defaults (if `autopopulate_in_workspace == True`).
        """
        new_runtime_envs = copy.deepcopy(runtime_envs)

        local_path_to_parsed_requirements: Dict[str, List[str]] = {}

        def _load_requirements_file_memoized(
            target: Union[str, List[str]]
        ) -> List[str]:
            if isinstance(target, list):
                return target
            elif target in local_path_to_parsed_requirements:
                return local_path_to_parsed_requirements[target]
            elif isinstance(target, str):
                parsed_requirements = parse_requirements_file(target)
                if parsed_requirements is None:
                    raise FileNotFoundError(
                        f"Requirements file {target} does not exist."
                    )
                local_path_to_parsed_requirements[target] = parsed_requirements
                return parsed_requirements
            else:
                raise TypeError("pip field in runtime_env must be a list or string.")

        for runtime_env in new_runtime_envs:
            if requirements_override is not None:
                # Explicitly-specified override from the user.
                runtime_env["pip"] = requirements_override
            elif (
                workspace_requirements_path is not None
                and "pip" not in runtime_env
                and "conda" not in runtime_env
            ):
                self.logger.info("Including workspace-managed pip dependencies.")
                runtime_env["pip"] = workspace_requirements_path

            if runtime_env.get("pip", None) is not None:
                # Load requirements from the file if necessary.
                runtime_env["pip"] = _load_requirements_file_memoized(
                    runtime_env["pip"]
                )

        return new_runtime_envs

    def override_and_upload_local_dirs(
        self,
        runtime_envs: List[Dict[str, Any]],
        *,
        working_dir_override: Optional[str],
        excludes_override: Optional[List[str]],
        cloud_id: Optional[str] = None,
        autopopulate_in_workspace: bool = True,
        additional_py_modules: Optional[List[str]] = None,
    ) -> List[Dict[str, Any]]:
        """Returns modified runtime_envs with all local dirs converted to remote URIs.

        The precedence for overrides is: explicit overrides passed in > fields in the existing
        runtime_envs > workspace defaults (if `autopopulate_in_workspace == True`).

        Each unique local directory across these fields will be uploaded once to cloud storage,
        then all occurrences of it in the config will be replaced with the corresponding remote URI.
        """
        new_runtime_envs = copy.deepcopy(runtime_envs)

        local_path_to_uri: Dict[str, str] = {}

        def _upload_dir_memoized(target: str, *, excludes: Optional[List[str]]) -> str:
            if is_dir_remote_uri(target):
                return target
            if target in local_path_to_uri:
                return local_path_to_uri[target]

            self.logger.info(f"Uploading local dir '{target}' to cloud storage.")
            uri = self._client.upload_local_dir_to_cloud_storage(
                target, cloud_id=cloud_id, excludes=excludes,
            )
            local_path_to_uri[target] = uri
            return uri

        for runtime_env in new_runtime_envs:
            # Extend, don't overwrite, excludes if it's provided.
            if excludes_override is not None:
                existing_excludes = runtime_env.get("excludes", None) or []
                runtime_env["excludes"] = existing_excludes + excludes_override

            final_excludes = runtime_env.get("excludes", [])

            new_working_dir = None
            if working_dir_override is not None:
                new_working_dir = working_dir_override
            elif "working_dir" in runtime_env:
                new_working_dir = runtime_env["working_dir"]
            elif autopopulate_in_workspace and self._client.inside_workspace():
                new_working_dir = "."

            if new_working_dir is not None:
                runtime_env["working_dir"] = _upload_dir_memoized(
                    new_working_dir, excludes=final_excludes
                )

            if additional_py_modules:
                existing_py_modules = runtime_env.get("py_modules", [])
                runtime_env["py_modules"] = existing_py_modules + additional_py_modules

            final_py_modules = runtime_env.get("py_modules", None)
            if final_py_modules is not None:
                runtime_env["py_modules"] = [
                    _upload_dir_memoized(py_module, excludes=final_excludes)
                    for py_module in final_py_modules
                ]

        return new_runtime_envs

    def resolve_compute_config(
        self, compute_config: Union[None, str, ComputeConfig]
    ) -> Tuple[str, str]:
        """Resolve the passed compute config to its ID and corresponding cloud ID.

        Accepts either:
            - A string of the form: '<name>[:<version>]'.
            - A dictionary from which an anonymous compute config will be built.

        Returns (compute_config_id, cloud_id).
        """
        if compute_config is None:
            compute_config_id = self._client.get_compute_config_id()
            assert compute_config_id is not None
        elif isinstance(compute_config, str):
            compute_config_id = self._client.get_compute_config_id(
                compute_config_name=compute_config,
            )
            if compute_config_id is None:
                raise ValueError(
                    f"The compute config '{compute_config}' does not exist."
                )
        else:
            _, compute_config_id = self._compute_config_sdk.create_compute_config(
                compute_config
            )

        return (
            compute_config_id,
            self.client.get_cloud_id(compute_config_id=compute_config_id),
        )

    def get_current_workspace_name(self) -> Optional[str]:
        """Get the name of the curernt workspace if running inside one."""
        if not self._client.inside_workspace():
            return None

        name = self._client.get_current_workspace_cluster().name
        # Defensively default to the workspace cluster name as-is if it doesn't
        # start with the expected prefix.
        if name.startswith(WORKSPACE_CLUSTER_NAME_PREFIX):
            name = name[len(WORKSPACE_CLUSTER_NAME_PREFIX) :]

        return name

    def get_containerfile_contents(self, path: str) -> str:
        """Get the full content of the containerfile as a string."""
        containerfile_path = pathlib.Path(path)
        if not containerfile_path.exists():
            raise FileNotFoundError(
                f"Containerfile '{containerfile_path}' does not exist."
            )
        if not containerfile_path.is_file():
            raise ValueError(f"Containerfile '{containerfile_path}' must be a file.")

        return containerfile_path.read_text()

    def _strip_none_values_from_dict(self, d: Dict) -> Dict:
        """Return a copy of the dictionary without any keys whose values are None.

        Recursively calls into any dictionary values.
        """
        result = {}
        for k, v in d.items():
            if isinstance(v, dict):
                result[k] = self._strip_none_values_from_dict(v)
            elif v is not None:
                result[k] = v

        return result

    def get_user_facing_compute_config(
        self, compute_config_id: str,
    ) -> Union[str, ComputeConfig]:
        """Get the compute config in a format to be displayed in a user-facing status.

        If the compute config refers to an anonymous compute config, its config
        object will be returned. Else the name of the compute config will be
        returned in the form: '<name>:<version>'.
        """
        compute_config = self._client.get_compute_config(compute_config_id)
        if compute_config is None:
            raise RuntimeError(
                f"Failed to get compute config for ID {compute_config_id}."
            )

        compute_config_name = compute_config.name
        if compute_config.version is not None:
            compute_config_name += f":{compute_config.version}"

        if not compute_config.anonymous:
            return compute_config_name

        return self._compute_config_sdk.get_compute_config(compute_config_name)



from typing import Optional, Union

from anyscale._private.sdk import sdk_command
from anyscale.service._private.service_sdk import ServiceSDK
from anyscale.service.models import ServiceConfig, ServiceState, ServiceStatus


_SERVICE_SDK_SINGLETON_KEY = "service_sdk"

_DEPLOY_EXAMPLE = """
import anyscale
from anyscale.service.models import ServiceConfig

anyscale.service.deploy(
    ServiceConfig(
        name="my-service",
        applications=[
            {"import_path": "main:app"},
        ],
        working_dir=".",
    ),
    canary_percent=50,
)
"""


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    ServiceSDK,
    doc_py_example=_DEPLOY_EXAMPLE,
    arg_docstrings={
        "config": "The config options defining the service.",
        "in_place": "Perform an in-place upgrade without starting a new cluster. This can be used for faster iteration during development but is *not* currently recommended for production deploys. This *cannot* be used to change cluster-level options such as image and compute config (they will be ignored).",
        "canary_percent": "The percentage of traffic to send to the canary version of the service (0-100). This can be used to manually shift traffic toward (or away from) the canary version. If not provided, traffic will be shifted incrementally toward the canary version until it reaches 100. Not supported when using --in-place.",
        "max_surge_percent": "Amount of excess capacity allowed to be used while updating the service (0-100). Defaults to 100. Not supported when using --in-place.",
    },
)
def deploy(
    config: ServiceConfig,
    *,
    in_place: bool = False,
    canary_percent: Optional[int] = None,
    max_surge_percent: Optional[int] = None,
    _sdk: ServiceSDK,
) -> str:
    """Deploy a service.

    If no service with the provided name is running, one will be created, else the existing service will be updated.

    This command is asynchronous, so it always returns immediately.

    Returns the id of the deployed service.
    """
    return _sdk.deploy(
        config,
        in_place=in_place,
        canary_percent=canary_percent,
        max_surge_percent=max_surge_percent,
    )


_ROLLBACK_EXAMPLE = """
import anyscale

anyscale.service.rollback(name="my-service")
"""


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    ServiceSDK,
    doc_py_example=_ROLLBACK_EXAMPLE,
    arg_docstrings={
        "name": "Name of the service. When running in a workspace, this defaults to the workspace name.",
        "max_surge_percent": "Amount of excess capacity allowed to be used while rolling back to the primary version of the service (0-100). Defaults to 100.",
    },
)
def rollback(
    name: Optional[str], *, max_surge_percent: Optional[int] = None, _sdk: ServiceSDK,
) -> str:
    """Rollback to the primary version of the service.

    This command can only be used when there is an active rollout in progress. The
    rollout will be cancelled and the service will revert to the primary version.

    This command is asynchronous, so it always returns immediately.

    Returns the id of the rolled back service.
    """
    return _sdk.rollback(name=name, max_surge_percent=max_surge_percent)


_TERMINATE_EXAMPLE = """
import anyscale

anyscale.service.terminate(name="my-service")
"""


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    ServiceSDK,
    doc_py_example=_TERMINATE_EXAMPLE,
    arg_docstrings={
        "name": "Name of the service. When running in a workspace, this defaults to the workspace name.",
    },
)
def terminate(name: Optional[str], *, _sdk: ServiceSDK) -> str:
    """Terminate a service.

    This command is asynchronous, so it always returns immediately.

    Returns the id of the terminated service.
    """
    return _sdk.terminate(name=name)


_STATUS_EXAMPLE = """
import anyscale
from anyscale.service.models import ServiceStatus

status: ServiceStatus = anyscale.service.status(name="my-service")
"""


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    ServiceSDK,
    doc_py_example=_STATUS_EXAMPLE,
    arg_docstrings={"name": "Name of the service.",},
)
def status(name: str, *, _sdk: ServiceSDK) -> ServiceStatus:
    """Get the status of a service."""
    return _sdk.status(name=name)


_WAIT_EXAMPLE = """
import anyscale
from anyscale.service.models import ServiceState

anyscale.service.wait(name="my-service", state=ServiceState.RUNNING)
"""


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    ServiceSDK,
    doc_py_example=_WAIT_EXAMPLE,
    arg_docstrings={
        "name": "Name of the service.",
        "state": "The state to wait for the service to reach.",
        "timeout_s": "Timeout to wait for the service to reach the target state.",
    },
)
def wait(
    name: str,
    *,
    state: Union[str, ServiceState] = ServiceState.RUNNING,
    timeout_s: float = 600,
    _sdk: ServiceSDK,
    _interval_s: float = 5,
):
    """Wait for a service to reach a target state."""
    _sdk.wait(
        name=name,
        state=ServiceState(state),
        timeout_s=timeout_s,
        interval_s=_interval_s,
    )



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class Cloud(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'str',
        'provider': 'CloudProviders',
        'region': 'str',
        'credentials': 'str',
        'config': 'CloudConfig',
        'is_k8s': 'bool',
        'is_aioa': 'bool',
        'availability_zones': 'list[str]',
        'is_bring_your_own_resource': 'bool',
        'is_private_cloud': 'bool',
        'cluster_management_stack_version': 'ClusterManagementStackVersions',
        'is_private_service_cloud': 'bool',
        'auto_add_user': 'bool',
        'id': 'str',
        'type': 'CloudTypes',
        'creator_id': 'str',
        'created_at': 'datetime',
        'status': 'CloudStatus',
        'state': 'CloudState',
        'version': 'CloudVersion',
        'is_default': 'bool',
        'customer_aggregated_logs_config_id': 'str'
    }

    attribute_map = {
        'name': 'name',
        'provider': 'provider',
        'region': 'region',
        'credentials': 'credentials',
        'config': 'config',
        'is_k8s': 'is_k8s',
        'is_aioa': 'is_aioa',
        'availability_zones': 'availability_zones',
        'is_bring_your_own_resource': 'is_bring_your_own_resource',
        'is_private_cloud': 'is_private_cloud',
        'cluster_management_stack_version': 'cluster_management_stack_version',
        'is_private_service_cloud': 'is_private_service_cloud',
        'auto_add_user': 'auto_add_user',
        'id': 'id',
        'type': 'type',
        'creator_id': 'creator_id',
        'created_at': 'created_at',
        'status': 'status',
        'state': 'state',
        'version': 'version',
        'is_default': 'is_default',
        'customer_aggregated_logs_config_id': 'customer_aggregated_logs_config_id'
    }

    def __init__(self, name=None, provider=None, region=None, credentials=None, config=None, is_k8s=False, is_aioa=False, availability_zones=None, is_bring_your_own_resource=None, is_private_cloud=False, cluster_management_stack_version=None, is_private_service_cloud=None, auto_add_user=False, id=None, type=None, creator_id=None, created_at=None, status=None, state=None, version=None, is_default=None, customer_aggregated_logs_config_id=None, local_vars_configuration=None):  # noqa: E501
        """Cloud - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._provider = None
        self._region = None
        self._credentials = None
        self._config = None
        self._is_k8s = None
        self._is_aioa = None
        self._availability_zones = None
        self._is_bring_your_own_resource = None
        self._is_private_cloud = None
        self._cluster_management_stack_version = None
        self._is_private_service_cloud = None
        self._auto_add_user = None
        self._id = None
        self._type = None
        self._creator_id = None
        self._created_at = None
        self._status = None
        self._state = None
        self._version = None
        self._is_default = None
        self._customer_aggregated_logs_config_id = None
        self.discriminator = None

        self.name = name
        self.provider = provider
        self.region = region
        self.credentials = credentials
        if config is not None:
            self.config = config
        if is_k8s is not None:
            self.is_k8s = is_k8s
        if is_aioa is not None:
            self.is_aioa = is_aioa
        if availability_zones is not None:
            self.availability_zones = availability_zones
        if is_bring_your_own_resource is not None:
            self.is_bring_your_own_resource = is_bring_your_own_resource
        if is_private_cloud is not None:
            self.is_private_cloud = is_private_cloud
        if cluster_management_stack_version is not None:
            self.cluster_management_stack_version = cluster_management_stack_version
        if is_private_service_cloud is not None:
            self.is_private_service_cloud = is_private_service_cloud
        if auto_add_user is not None:
            self.auto_add_user = auto_add_user
        self.id = id
        self.type = type
        self.creator_id = creator_id
        self.created_at = created_at
        if status is not None:
            self.status = status
        if state is not None:
            self.state = state
        if version is not None:
            self.version = version
        self.is_default = is_default
        self.customer_aggregated_logs_config_id = customer_aggregated_logs_config_id

    @property
    def name(self):
        """Gets the name of this Cloud.  # noqa: E501

        Name of this Cloud.  # noqa: E501

        :return: The name of this Cloud.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this Cloud.

        Name of this Cloud.  # noqa: E501

        :param name: The name of this Cloud.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def provider(self):
        """Gets the provider of this Cloud.  # noqa: E501

        Provider of this Cloud (e.g. AWS).  # noqa: E501

        :return: The provider of this Cloud.  # noqa: E501
        :rtype: CloudProviders
        """
        return self._provider

    @provider.setter
    def provider(self, provider):
        """Sets the provider of this Cloud.

        Provider of this Cloud (e.g. AWS).  # noqa: E501

        :param provider: The provider of this Cloud.  # noqa: E501
        :type: CloudProviders
        """
        if self.local_vars_configuration.client_side_validation and provider is None:  # noqa: E501
            raise ValueError("Invalid value for `provider`, must not be `None`")  # noqa: E501

        self._provider = provider

    @property
    def region(self):
        """Gets the region of this Cloud.  # noqa: E501

        Region this Cloud is operating in. This value needs to be supported by this Cloud's provider. (e.g. us-west-2)  # noqa: E501

        :return: The region of this Cloud.  # noqa: E501
        :rtype: str
        """
        return self._region

    @region.setter
    def region(self, region):
        """Sets the region of this Cloud.

        Region this Cloud is operating in. This value needs to be supported by this Cloud's provider. (e.g. us-west-2)  # noqa: E501

        :param region: The region of this Cloud.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and region is None:  # noqa: E501
            raise ValueError("Invalid value for `region`, must not be `None`")  # noqa: E501

        self._region = region

    @property
    def credentials(self):
        """Gets the credentials of this Cloud.  # noqa: E501

        Credentials needed to interact with this Cloud.  # noqa: E501

        :return: The credentials of this Cloud.  # noqa: E501
        :rtype: str
        """
        return self._credentials

    @credentials.setter
    def credentials(self, credentials):
        """Sets the credentials of this Cloud.

        Credentials needed to interact with this Cloud.  # noqa: E501

        :param credentials: The credentials of this Cloud.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and credentials is None:  # noqa: E501
            raise ValueError("Invalid value for `credentials`, must not be `None`")  # noqa: E501

        self._credentials = credentials

    @property
    def config(self):
        """Gets the config of this Cloud.  # noqa: E501

        Additional configurable properties of this Cloud.  # noqa: E501

        :return: The config of this Cloud.  # noqa: E501
        :rtype: CloudConfig
        """
        return self._config

    @config.setter
    def config(self, config):
        """Sets the config of this Cloud.

        Additional configurable properties of this Cloud.  # noqa: E501

        :param config: The config of this Cloud.  # noqa: E501
        :type: CloudConfig
        """

        self._config = config

    @property
    def is_k8s(self):
        """Gets the is_k8s of this Cloud.  # noqa: E501

        Whether this cloud is managed via Kubernetes.  # noqa: E501

        :return: The is_k8s of this Cloud.  # noqa: E501
        :rtype: bool
        """
        return self._is_k8s

    @is_k8s.setter
    def is_k8s(self, is_k8s):
        """Sets the is_k8s of this Cloud.

        Whether this cloud is managed via Kubernetes.  # noqa: E501

        :param is_k8s: The is_k8s of this Cloud.  # noqa: E501
        :type: bool
        """

        self._is_k8s = is_k8s

    @property
    def is_aioa(self):
        """Gets the is_aioa of this Cloud.  # noqa: E501

        Whether this cloud is an AIOA cloud.  # noqa: E501

        :return: The is_aioa of this Cloud.  # noqa: E501
        :rtype: bool
        """
        return self._is_aioa

    @is_aioa.setter
    def is_aioa(self, is_aioa):
        """Sets the is_aioa of this Cloud.

        Whether this cloud is an AIOA cloud.  # noqa: E501

        :param is_aioa: The is_aioa of this Cloud.  # noqa: E501
        :type: bool
        """

        self._is_aioa = is_aioa

    @property
    def availability_zones(self):
        """Gets the availability_zones of this Cloud.  # noqa: E501

        The availability zones that instances of this cloud are allowed to be launched in.  # noqa: E501

        :return: The availability_zones of this Cloud.  # noqa: E501
        :rtype: list[str]
        """
        return self._availability_zones

    @availability_zones.setter
    def availability_zones(self, availability_zones):
        """Sets the availability_zones of this Cloud.

        The availability zones that instances of this cloud are allowed to be launched in.  # noqa: E501

        :param availability_zones: The availability_zones of this Cloud.  # noqa: E501
        :type: list[str]
        """

        self._availability_zones = availability_zones

    @property
    def is_bring_your_own_resource(self):
        """Gets the is_bring_your_own_resource of this Cloud.  # noqa: E501

        Whether the resources of this cloud are provided by the customer.  # noqa: E501

        :return: The is_bring_your_own_resource of this Cloud.  # noqa: E501
        :rtype: bool
        """
        return self._is_bring_your_own_resource

    @is_bring_your_own_resource.setter
    def is_bring_your_own_resource(self, is_bring_your_own_resource):
        """Sets the is_bring_your_own_resource of this Cloud.

        Whether the resources of this cloud are provided by the customer.  # noqa: E501

        :param is_bring_your_own_resource: The is_bring_your_own_resource of this Cloud.  # noqa: E501
        :type: bool
        """

        self._is_bring_your_own_resource = is_bring_your_own_resource

    @property
    def is_private_cloud(self):
        """Gets the is_private_cloud of this Cloud.  # noqa: E501

        Whether this cloud is a private cloud.  # noqa: E501

        :return: The is_private_cloud of this Cloud.  # noqa: E501
        :rtype: bool
        """
        return self._is_private_cloud

    @is_private_cloud.setter
    def is_private_cloud(self, is_private_cloud):
        """Sets the is_private_cloud of this Cloud.

        Whether this cloud is a private cloud.  # noqa: E501

        :param is_private_cloud: The is_private_cloud of this Cloud.  # noqa: E501
        :type: bool
        """

        self._is_private_cloud = is_private_cloud

    @property
    def cluster_management_stack_version(self):
        """Gets the cluster_management_stack_version of this Cloud.  # noqa: E501

        The cluster management stack version of the cloud.  # noqa: E501

        :return: The cluster_management_stack_version of this Cloud.  # noqa: E501
        :rtype: ClusterManagementStackVersions
        """
        return self._cluster_management_stack_version

    @cluster_management_stack_version.setter
    def cluster_management_stack_version(self, cluster_management_stack_version):
        """Sets the cluster_management_stack_version of this Cloud.

        The cluster management stack version of the cloud.  # noqa: E501

        :param cluster_management_stack_version: The cluster_management_stack_version of this Cloud.  # noqa: E501
        :type: ClusterManagementStackVersions
        """

        self._cluster_management_stack_version = cluster_management_stack_version

    @property
    def is_private_service_cloud(self):
        """Gets the is_private_service_cloud of this Cloud.  # noqa: E501

        Whether services created in this cloud should be private.  # noqa: E501

        :return: The is_private_service_cloud of this Cloud.  # noqa: E501
        :rtype: bool
        """
        return self._is_private_service_cloud

    @is_private_service_cloud.setter
    def is_private_service_cloud(self, is_private_service_cloud):
        """Sets the is_private_service_cloud of this Cloud.

        Whether services created in this cloud should be private.  # noqa: E501

        :param is_private_service_cloud: The is_private_service_cloud of this Cloud.  # noqa: E501
        :type: bool
        """

        self._is_private_service_cloud = is_private_service_cloud

    @property
    def auto_add_user(self):
        """Gets the auto_add_user of this Cloud.  # noqa: E501

        Whether all users in the organization should be automatically added to this cloud. This field is only relevant for organizations with cloud isolation enabled, because all users in the organization automatically have access to all clouds if cloud isolation is not enabled.  # noqa: E501

        :return: The auto_add_user of this Cloud.  # noqa: E501
        :rtype: bool
        """
        return self._auto_add_user

    @auto_add_user.setter
    def auto_add_user(self, auto_add_user):
        """Sets the auto_add_user of this Cloud.

        Whether all users in the organization should be automatically added to this cloud. This field is only relevant for organizations with cloud isolation enabled, because all users in the organization automatically have access to all clouds if cloud isolation is not enabled.  # noqa: E501

        :param auto_add_user: The auto_add_user of this Cloud.  # noqa: E501
        :type: bool
        """

        self._auto_add_user = auto_add_user

    @property
    def id(self):
        """Gets the id of this Cloud.  # noqa: E501

        Server assigned unique identifier.  # noqa: E501

        :return: The id of this Cloud.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this Cloud.

        Server assigned unique identifier.  # noqa: E501

        :param id: The id of this Cloud.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def type(self):
        """Gets the type of this Cloud.  # noqa: E501


        :return: The type of this Cloud.  # noqa: E501
        :rtype: CloudTypes
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this Cloud.


        :param type: The type of this Cloud.  # noqa: E501
        :type: CloudTypes
        """
        if self.local_vars_configuration.client_side_validation and type is None:  # noqa: E501
            raise ValueError("Invalid value for `type`, must not be `None`")  # noqa: E501

        self._type = type

    @property
    def creator_id(self):
        """Gets the creator_id of this Cloud.  # noqa: E501

        ID of the User who created this Cloud.  # noqa: E501

        :return: The creator_id of this Cloud.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this Cloud.

        ID of the User who created this Cloud.  # noqa: E501

        :param creator_id: The creator_id of this Cloud.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_id is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_id`, must not be `None`")  # noqa: E501

        self._creator_id = creator_id

    @property
    def created_at(self):
        """Gets the created_at of this Cloud.  # noqa: E501

        Time when this Cloud was created.  # noqa: E501

        :return: The created_at of this Cloud.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this Cloud.

        Time when this Cloud was created.  # noqa: E501

        :param created_at: The created_at of this Cloud.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def status(self):
        """Gets the status of this Cloud.  # noqa: E501

        The status of this cloud.  # noqa: E501

        :return: The status of this Cloud.  # noqa: E501
        :rtype: CloudStatus
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this Cloud.

        The status of this cloud.  # noqa: E501

        :param status: The status of this Cloud.  # noqa: E501
        :type: CloudStatus
        """

        self._status = status

    @property
    def state(self):
        """Gets the state of this Cloud.  # noqa: E501

        The state of this cloud.  # noqa: E501

        :return: The state of this Cloud.  # noqa: E501
        :rtype: CloudState
        """
        return self._state

    @state.setter
    def state(self, state):
        """Sets the state of this Cloud.

        The state of this cloud.  # noqa: E501

        :param state: The state of this Cloud.  # noqa: E501
        :type: CloudState
        """

        self._state = state

    @property
    def version(self):
        """Gets the version of this Cloud.  # noqa: E501

        The version of the cloud.  # noqa: E501

        :return: The version of this Cloud.  # noqa: E501
        :rtype: CloudVersion
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this Cloud.

        The version of the cloud.  # noqa: E501

        :param version: The version of this Cloud.  # noqa: E501
        :type: CloudVersion
        """

        self._version = version

    @property
    def is_default(self):
        """Gets the is_default of this Cloud.  # noqa: E501

        Whether this cloud is the default cloud.  # noqa: E501

        :return: The is_default of this Cloud.  # noqa: E501
        :rtype: bool
        """
        return self._is_default

    @is_default.setter
    def is_default(self, is_default):
        """Sets the is_default of this Cloud.

        Whether this cloud is the default cloud.  # noqa: E501

        :param is_default: The is_default of this Cloud.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and is_default is None:  # noqa: E501
            raise ValueError("Invalid value for `is_default`, must not be `None`")  # noqa: E501

        self._is_default = is_default

    @property
    def customer_aggregated_logs_config_id(self):
        """Gets the customer_aggregated_logs_config_id of this Cloud.  # noqa: E501

        the id of the customer aggregated logs config associated with this cloud.  # noqa: E501

        :return: The customer_aggregated_logs_config_id of this Cloud.  # noqa: E501
        :rtype: str
        """
        return self._customer_aggregated_logs_config_id

    @customer_aggregated_logs_config_id.setter
    def customer_aggregated_logs_config_id(self, customer_aggregated_logs_config_id):
        """Sets the customer_aggregated_logs_config_id of this Cloud.

        the id of the customer aggregated logs config associated with this cloud.  # noqa: E501

        :param customer_aggregated_logs_config_id: The customer_aggregated_logs_config_id of this Cloud.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and customer_aggregated_logs_config_id is None:  # noqa: E501
            raise ValueError("Invalid value for `customer_aggregated_logs_config_id`, must not be `None`")  # noqa: E501

        self._customer_aggregated_logs_config_id = customer_aggregated_logs_config_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Cloud):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Cloud):
            return True

        return self.to_dict() != other.to_dict()



from anyscale._private.anyscale_client.anyscale_client import AnyscaleClient
from anyscale._private.anyscale_client.common import (
    AnyscaleClientInterface,
    DEFAULT_PYTHON_VERSION,
    DEFAULT_RAY_VERSION,
    WORKSPACE_CLUSTER_NAME_PREFIX,
)
from anyscale._private.anyscale_client.fake_anyscale_client import FakeAnyscaleClient



import os
from typing import Any, Dict, List, Optional

import click
import tabulate
import yaml

from anyscale import AnyscaleSDK
from anyscale.cli_logger import BlockLogger

# There are two API clients: one that's exposed to customers via the SDK (ExternalApi)
# and one that's only used internally by the UI and CLI (InternalApi).
from anyscale.client.openapi_client.api.default_api import DefaultApi as InternalApi
from anyscale.client.openapi_client.models.decorated_list_service_api_model import (
    DecoratedListServiceAPIModel,
)
from anyscale.client.openapi_client.models.decorated_production_service_v2_api_model import (
    DecoratedProductionServiceV2APIModel,
)
from anyscale.controllers.base_controller import BaseController
from anyscale.models.service_model import ServiceConfig
from anyscale.project import infer_project_id
from anyscale.sdk.anyscale_client.models import (
    ApplyServiceModel,
    RollbackServiceModel,
    ServiceModel,
)
from anyscale.tables import ServicesTable
from anyscale.util import (
    get_endpoint,
    populate_unspecified_cluster_configs_from_current_workspace,
)
from anyscale.utils.runtime_env import override_runtime_env_config
from anyscale.utils.workload_types import Workload
from anyscale.utils.workspace_notification import (
    send_workspace_notification,
    WorkspaceNotification,
    WorkspaceNotificationAction,
)


class ServiceController(BaseController):
    def __init__(self, *, sdk: Optional[AnyscaleSDK] = None):
        super().__init__()
        self.log = BlockLogger()
        self.sdk = sdk if sdk is not None else AnyscaleSDK()

    @property
    def internal_api_client(self) -> InternalApi:
        return self.api_client

    def _format_apply_service_model(self, config: ServiceConfig) -> ApplyServiceModel:
        if not config.ray_serve_config:
            raise click.ClickException(
                "ray_serve_config must be provided in the service configuration."
            )

        if "import_path" in config.ray_serve_config:
            self.log.warning(
                "The single application Ray Serve config is deprecated in Ray 2.8 and "
                "removed in Ray 2.9. Please migrate to the multi-application Ray Serve "
                "config format. Please see "
                "https://docs.ray.io/en/latest/serve/multi-app.html#migrating-from-a-single-application-config "
                "for migration instructions."
            )

        return ApplyServiceModel(
            name=config.name,
            description=config.description or "Service updated from CLI",
            project_id=config.project_id,
            version=config.version,
            canary_percent=config.canary_percent,
            ray_serve_config=config.ray_serve_config,
            ray_gcs_external_storage_config=config.ray_gcs_external_storage_config,
            build_id=config.build_id,
            compute_config_id=config.compute_config_id,
            rollout_strategy=config.rollout_strategy,
            config=config.config,
            auto_complete_rollout=config.auto_complete_rollout,
            max_surge_percent=config.max_surge_percent,
        )

    def _get_services_by_name(
        self,
        *,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        created_by_me: bool,
        max_items: int,
    ) -> List[DecoratedListServiceAPIModel]:
        """Makes an API call to get all services matching the provided filters.

        Note that "name" is an *exact match* (different from the REST API semantics).
        """
        creator_id = None
        if created_by_me:
            user_info_response = (
                self.internal_api_client.get_user_info_api_v2_userinfo_get()
            )
            creator_id = user_info_response.result.id

        paging_token = None
        services_list: List[DecoratedListServiceAPIModel] = []
        while len(services_list) < max_items:
            resp = self.internal_api_client.list_services_api_v2_services_v2_get(
                project_id=project_id,
                name=name,
                creator_id=creator_id,
                count=10,
                paging_token=paging_token,
            )
            for r in resp.results:
                # The 'name' filter in the list endpoint is not exact match.
                if name is None or r.name == name:
                    services_list.append(r)
                if len(services_list) >= max_items:
                    break

            paging_token = resp.metadata.next_paging_token
            if paging_token is None:
                break

        return services_list[:max_items]

    def _get_service_id_from_name(
        self, service_name: str, project_id: Optional[str]
    ) -> str:
        """Get the ID for a service by name.

        If project_id is specified, filter to that project, else don't filter on project_id and
        instead error if there are multiple services with the name.

        Raises an exception if there are zero or multiple services with the given name.
        """
        results = self._get_services_by_name(
            name=service_name, project_id=project_id, created_by_me=False, max_items=10
        )

        if len(results) == 0:
            raise click.ClickException(
                f"No service with name '{service_name}' was found. "
                "Please verify that this service exists and you have access to it."
            )
        elif len(results) > 1:
            raise click.ClickException(
                f"There are multiple services with name '{service_name}'. "
                "Please filter using --project-id or specify the --service-id instead. "
                f"Services found: \n{ServicesTable(results)}"
            )

        return results[0].id

    def get_service_id(
        self,
        *,
        service_id: Optional[str] = None,
        service_name: Optional[str] = None,
        service_config_file: Optional[str] = None,
        project_id: Optional[str] = None,
    ) -> str:
        """Get the service ID given the ID, name, or config file.

        This is a utility used by multiple CLI commands to standardize mapping these options
        to a service_id.

        The precedence is: service_id > service_name > service_config_file.

        If the project_id is specified directly or via the service_config_file it will be used
        to filter the query. Else we will try to find the service across all projects and error
        if there are multiple.
        """
        if service_id is not None:
            if service_name is not None or service_config_file is not None:
                raise click.ClickException(
                    "Only one of service ID, name, or config file should be specified."
                )
        elif service_name is not None:
            if service_config_file is not None:
                raise click.ClickException(
                    "Only one of service ID, name, or config file should be specified."
                )
            service_id = self._get_service_id_from_name(service_name, project_id)
        elif service_config_file is not None:
            service_config: ServiceConfig = self.parse_service_config_dict(
                self.read_service_config_file(service_config_file)
            )
            # Allow the passed project_id to override the one in the file.
            project_id = project_id or service_config.project_id
            service_id = self._get_service_id_from_name(service_config.name, project_id)
        else:
            raise click.ClickException(
                "Service ID, name, or config file must be specified."
            )

        return service_id

    def override_config_options(
        self,
        service_config: ServiceConfig,
        *,
        auto_complete_rollout: bool,
        name: Optional[str] = None,
        description: Optional[str] = None,
        version: Optional[str] = None,
        max_surge_percent: Optional[int] = None,
        canary_percent: Optional[int] = None,
        rollout_strategy: Optional[str] = None,
    ) -> ServiceConfig:
        """
        Given the input config and user overrides, this method
        constructs the expected service configuration.

        Note that if the command is run from a workspace, some values
        may be overwritten.
        """
        if name:
            service_config.name = name

        if description:
            service_config.description = description

        if version:
            service_config.version = version

        if canary_percent is not None:
            service_config.canary_percent = canary_percent

        if rollout_strategy:
            service_config.rollout_strategy = rollout_strategy

        if service_config.rollout_strategy:
            service_config.rollout_strategy = service_config.rollout_strategy.upper()

        if auto_complete_rollout is not None:
            service_config.auto_complete_rollout = auto_complete_rollout

        if max_surge_percent is not None:
            service_config.max_surge_percent = max_surge_percent

        if (
            service_config.max_surge_percent is not None
            and service_config.rollout_strategy == "IN_PLACE"
        ):
            raise ValueError(
                "--max-surge-percent is not supported for IN_PLACE rollouts."
            )

        return service_config

    def read_service_config_file(self, service_config_file: str) -> Dict[str, Any]:
        if not os.path.exists(service_config_file):
            raise click.ClickException(f"Config file {service_config_file} not found.")

        with open(service_config_file) as f:
            return yaml.safe_load(f)

    def parse_service_config_dict(self, config_dict: Dict[str, Any]) -> ServiceConfig:
        # If running in a workspace, auto-populate unspecified fields.
        config_dict = populate_unspecified_cluster_configs_from_current_workspace(
            config_dict, self.sdk, populate_name=True
        )
        return ServiceConfig.parse_obj(config_dict)

    def rollout(  # noqa: PLR0913
        self,
        config_dict: Dict[str, Any],
        *,
        auto_complete_rollout: bool,
        name: Optional[str] = None,
        version: Optional[str] = None,
        max_surge_percent: Optional[int] = None,
        canary_percent: Optional[int] = None,
        rollout_strategy: Optional[str] = None,
    ):
        """
        Deploys a Service 2.0.
        """
        config = self.override_config_options(
            self.parse_service_config_dict(config_dict),
            name=name,
            version=version,
            max_surge_percent=max_surge_percent,
            canary_percent=canary_percent,
            rollout_strategy=rollout_strategy,
            auto_complete_rollout=auto_complete_rollout,
        )

        config.project_id = infer_project_id(
            self.sdk,
            self.internal_api_client,
            self.log,
            project_id=config.project_id,
            cluster_compute_id=config.compute_config_id,
            cluster_compute=config.compute_config,
            cloud=config.cloud,
        )

        service_config = config.config
        if service_config:
            access_config = service_config.get("access", {})

            if (
                isinstance(access_config, dict)
                and access_config.get("use_bearer_token", True) is False
            ):
                self.log.warning(
                    "Using bearer token for authorization is disabled: no authorization guarding will be used for this service."
                )

        if not config.ray_serve_config:
            config.ray_serve_config = {}

        self._overwrite_runtime_env_in_v2_ray_serve_config(config)

        apply_service_model: ApplyServiceModel = self._format_apply_service_model(
            config
        )
        service: ServiceModel = self.sdk.rollout_service(apply_service_model).result
        self.log.info(f"Service {service.id} rollout initiated.")
        self.log.info(
            f'View the service in the UI at {get_endpoint(f"/services/{service.id}")}'
        )

        url = service.base_url
        auth_token = service.auth_token
        self.log.info(
            "You can query the service endpoint using the curl request below:"
        )
        self.log.info(f"curl -H 'Authorization: Bearer {auth_token}' {url}")

        send_workspace_notification(
            self.sdk,
            WorkspaceNotification(
                body=f"Service {service.name} rollout initiated.",
                action=WorkspaceNotificationAction(
                    type="navigate-service", title="View Service", value=service.id,
                ),
            ),
        )

    def _overwrite_runtime_env_in_v2_ray_serve_config(self, config: ServiceConfig):
        """Modifies config in place."""
        ray_serve_config = config.ray_serve_config
        if "applications" in ray_serve_config:
            for ray_serve_app_config in ray_serve_config["applications"]:
                ray_serve_app_config["runtime_env"] = override_runtime_env_config(
                    runtime_env=ray_serve_app_config.get("runtime_env"),
                    anyscale_api_client=self.sdk,
                    api_client=self.internal_api_client,
                    workload_type=Workload.SERVICES,
                    compute_config_id=config.compute_config_id,
                    log=self.log,
                )

        else:
            ray_serve_config["runtime_env"] = override_runtime_env_config(
                runtime_env=ray_serve_config.get("runtime_env"),
                anyscale_api_client=self.sdk,
                api_client=self.internal_api_client,
                workload_type=Workload.SERVICES,
                compute_config_id=config.compute_config_id,
                log=self.log,
            )

    def list(
        self,
        *,
        name: Optional[str] = None,
        service_id: Optional[str] = None,
        project_id: Optional[str] = None,
        created_by_me: bool = False,
        max_items: int = 10,
    ) -> None:
        self._list_via_service_list_endpoint(
            name=name,
            service_id=service_id,
            project_id=project_id,
            created_by_me=created_by_me,
            max_items=max_items,
        )

    def _list_via_service_list_endpoint(
        self,
        *,
        name: Optional[str] = None,
        service_id: Optional[str] = None,
        project_id: Optional[str] = None,
        created_by_me: bool = False,
        max_items: int,
    ):
        services = []
        if service_id:
            service_v2_result: DecoratedProductionServiceV2APIModel = (
                self.internal_api_client.get_service_api_v2_services_v2_service_id_get(
                    service_id
                ).result
            )
            services.append(
                [
                    service_v2_result.name,
                    service_v2_result.id,
                    service_v2_result.current_state,
                    # TODO: change to email once https://github.com/anyscale/product/pull/18189 is merged
                    service_v2_result.creator_id,
                ]
            )

        else:
            services_data = self._get_services_by_name(
                name=name,
                project_id=project_id,
                created_by_me=created_by_me,
                max_items=max_items,
            )

            services = [
                [
                    service.name,
                    service.id,
                    service.current_state,
                    service.creator.email,
                ]
                for service in services_data
            ]

        table = tabulate.tabulate(
            services,
            headers=["NAME", "ID", "CURRENT STATE", "CREATOR"],
            tablefmt="plain",
        )
        self.log.info(f'View your Services in the UI at {get_endpoint("/services")}')
        self.log.info(f"Services:\n{table}")

    def rollback(self, service_id: str, max_surge_percent: Optional[int] = None):
        service: ServiceModel = self.sdk.rollback_service(
            service_id,
            rollback_service_model=RollbackServiceModel(
                max_surge_percent=max_surge_percent
            ),
        ).result

        self.log.info(f"Service {service.id} rollback initiated.")
        self.log.info(
            f'View the service in the UI at {get_endpoint(f"/services/{service.id}")}'
        )

    def terminate(self, service_id: str):
        service: ServiceModel = self.sdk.terminate_service(service_id).result
        self.log.info(f"Service {service.id} terminate initiated.")
        self.log.info(
            f'View the service in the UI at {get_endpoint(f"/services/{service.id}")}'
        )



import re

import pytest

from anyscale._private.models.image_uri import ImageURI


class TestImageURI:
    @pytest.mark.parametrize(
        ("image_uri_str", "expected_name"),
        [
            ("docker.us.com/my/fakeimage:latest", "docker-us-com-my-fakeimage-latest"),
            ("library/ubuntu@sha256:45b23dee08", "library-ubuntu-sha256-45b23dee08"),
        ],
    )
    def test_image_uri_to_cluster_env_name(
        self, image_uri_str, expected_name,
    ):
        image_uri = ImageURI.from_str(image_uri_str)
        assert image_uri.to_cluster_env_name() == expected_name

    def test_empty_image_uri(self,):
        with pytest.raises(ValueError, match="image_uri cannot be empty."):
            ImageURI.from_str("")

    @pytest.mark.parametrize(
        ("image_uri_str", "is_legacy_cluster_env_image"),
        [("docker.us.com/my/fakeimage:laest", False), ("anyscale/image/name:5", True),],
    )
    def test_image_uri_is_legacy_cluster_env_image(
        self, image_uri_str, is_legacy_cluster_env_image
    ):
        assert (
            ImageURI.from_str(image_uri_str).is_cluster_env_image()
            == is_legacy_cluster_env_image
        )

    @pytest.mark.parametrize(
        ("image_uri", "valid"),
        [
            ("anyscale/cluster_env/default_cluster_env_2.9.3_py39:1", True),
            ("docker.io/libaray/ubuntu:latest", True),
            ("ubuntu:latest", True),
            ("python:3.8", True),
            ("myregistry.local:5000/testing/test-image:1.0.0", True),
            ("localhost:5000/myusername/myrepository:latest", True),
            ("localhost:5000/myusername/my/repository:latest", True),
            ("valid/withouttag", True),
            ("valid_name/withtag_and_digest:v2@sha213", True),
            ("valid_name/withtag_and_digest@sha213", True),
            ("valid_name/withtag_and_digest:@sha213", False),
            ("http://myregistry.local:5000/testing/test-image:1.0.0", False),
            (
                "us-docker.pkg.dev/anyscale-workspace-templates/workspace-templates/rag-dev-bootcamp-mar-2024:raynightly-py310",
                True,
            ),
            (
                "241673607239.dkr.ecr.us-west-2.amazonaws.com/aviary:release-endpoints_aica-e8873f9361e29289bcea2de47d967bdd5291ac46",
                True,
            ),
        ],
    )
    def test_image_uri_validation(self, image_uri: str, valid: bool):
        if valid:
            assert str(ImageURI.from_str(image_uri)) == image_uri
        else:
            with pytest.raises(
                ValueError,
                match=re.escape(
                    f"Invalid image URI: '{image_uri}'. Must be in the format: '[registry_host/]user_name/repository[:tag][@digest]'."
                ),
            ):
                ImageURI.from_str(image_uri)



import json
import tempfile
from unittest.mock import Mock, patch

import click
import pytest

import anyscale
from anyscale.authenticate import AuthenticationBlock
from anyscale.client.openapi_client.rest import ApiException as ApiExceptionInternal
import anyscale.conf


def test_load_credentials_env_var(monkeypatch):
    """
    Test credentials loaded from environment variable even when credentials exist in file.
    """
    with patch.multiple(
        "anyscale.authenticate.AuthenticationBlock", __init__=Mock(return_value=None)
    ):
        auth_api_client = AuthenticationBlock()

    monkeypatch.setenv("ANYSCALE_CLI_TOKEN", "sss_os_environ")
    with tempfile.NamedTemporaryFile("w") as temp_credentials_file:
        temp_credentials_file.write(json.dumps({"cli_token": "sss_file_credential"}))
        temp_credentials_file.flush()
        anyscale.authenticate.CREDENTIALS_FILE = temp_credentials_file.name
        assert auth_api_client._load_credentials() == (
            "sss_os_environ",
            "ANYSCALE_CLI_TOKEN",
        )


def test_load_credentials_file(monkeypatch):
    """
    Test credentials loaded from file when credentials don't exist in environment variable.
    """
    with patch.multiple(
        "anyscale.authenticate.AuthenticationBlock", __init__=Mock(return_value=None)
    ):
        auth_api_client = AuthenticationBlock()

    monkeypatch.delenv("ANYSCALE_CLI_TOKEN", raising=False)
    with tempfile.NamedTemporaryFile("w") as temp_credentials_file:
        temp_credentials_file.write(json.dumps({"cli_token": "sss_file_credential"}))
        temp_credentials_file.flush()
        anyscale.authenticate.CREDENTIALS_FILE = temp_credentials_file.name
        assert auth_api_client._load_credentials() == (
            "sss_file_credential",
            anyscale.authenticate.CREDENTIALS_FILE,
        )


def test_validate_credentials_format() -> None:
    """
    Test credentials are of the correct format.
    """
    with patch.multiple(
        "anyscale.authenticate.AuthenticationBlock", __init__=Mock(return_value=None)
    ):
        auth_api_client = AuthenticationBlock()

    with pytest.raises(click.ClickException):
        auth_api_client._validate_credentials_format("abc")

    with pytest.raises(click.ClickException):
        auth_api_client._validate_credentials_format("")

    # Should not raise an exception.
    auth_api_client._validate_credentials_format("sss_abcdefghjijklmnop")

    # Possible format of new tokens.
    auth_api_client._validate_credentials_format("ast_potential_new_token")


def test_validate_api_client_auth() -> None:
    """
    Test credentials are validated with by calling /api/v2/userinfo.
    """
    with patch.multiple(
        "anyscale.authenticate.AuthenticationBlock", __init__=Mock(return_value=None)
    ):
        auth_api_client = AuthenticationBlock()

    # Test credentials are validated when /api/v2/userinfo returns successful response.
    auth_api_client.api_client = Mock()
    auth_api_client._validate_api_client_auth()

    # Test error is raised when /api/v2/userinfo returns 401 not authenticated response.
    auth_api_client.api_client.get_user_info_api_v2_userinfo_get = Mock(
        side_effect=ApiExceptionInternal(status=401)
    )
    with pytest.raises(click.ClickException):
        auth_api_client._validate_api_client_auth()


def test_warn_credential_file_permissions():
    """
    Test if logger warns when credentials file permissions are too open
    """
    with patch.multiple(
        "os.path",
        expanduser=Mock(return_value="/home/usr0/.anyscale/credentials.json"),
        exists=Mock(return_value=True),
    ), patch("os.stat", Mock()):
        mock_log_warning = Mock()

        nowarn_modes = [0o600, 0o700, 0o400]
        for mode in nowarn_modes:
            with patch("stat.S_IMODE", Mock(return_value=mode)), patch(
                "anyscale.cli_logger.BlockLogger.warning", mock_log_warning
            ):
                auth = AuthenticationBlock(cli_token="", validate_credentials=False)
                auth._warn_credential_file_permissions(
                    "/home/usr0/.anyscale/credentials.json"
                )
            mock_log_warning.assert_not_called()
            mock_log_warning.reset_mock()

        warn_modes = [0o644, 0o777, 0o640, 0o770]
        for mode in warn_modes:
            with patch("stat.S_IMODE", Mock(return_value=mode)), patch(
                "anyscale.cli_logger.BlockLogger.warning", mock_log_warning
            ):
                auth = AuthenticationBlock(cli_token="", validate_credentials=False)
                auth._warn_credential_file_permissions(
                    "/home/usr0/.anyscale/credentials.json"
                )
            mock_log_warning.assert_called()
            mock_log_warning.reset_mock()


def test_warn_credential_file_permissions_dont_on_windows():
    """
    Test that no warning is printed on Windows
    """
    mock_log_warning = Mock()
    with patch.multiple(
        "os.path",
        expanduser=Mock(return_value="/home/usr0/.anyscale/credentials.json"),
        exists=Mock(return_value=True),
    ), patch("os.stat", Mock()), patch(
        "anyscale.cli_logger.BlockLogger.warning", mock_log_warning
    ), patch(
        "platform.system", Mock(return_value="Windows")
    ):
        auth = AuthenticationBlock(cli_token="", validate_credentials=False)
        auth._warn_credential_file_permissions("/home/usr0/.anyscale/credentials.json")

    mock_log_warning.assert_not_called()



from typing import Optional

from anyscale._private.anyscale_client import (
    AnyscaleClient,
    AnyscaleClientInterface,
)
from anyscale._private.sdk.timer import RealTimer, Timer
from anyscale.cli_logger import BlockLogger


class BaseSDK:
    """Shared parent class for all SDKs."""

    def __init__(
        self,
        *,
        logger: Optional[BlockLogger] = None,
        client: Optional[AnyscaleClientInterface] = None,
        timer: Optional[Timer] = None,
    ):
        self._logger = logger or BlockLogger()
        self._client = client or AnyscaleClient()
        self._timer = timer or RealTimer()

    @property
    def logger(self) -> BlockLogger:
        return self._logger

    @property
    def client(self) -> AnyscaleClientInterface:
        return self._client

    @property
    def timer(self) -> Timer:
        return self._timer



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class UpdateOrganization(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'sso_mode': 'SSOMode'
    }

    attribute_map = {
        'sso_mode': 'sso_mode'
    }

    def __init__(self, sso_mode=None, local_vars_configuration=None):  # noqa: E501
        """UpdateOrganization - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._sso_mode = None
        self.discriminator = None

        self.sso_mode = sso_mode

    @property
    def sso_mode(self):
        """Gets the sso_mode of this UpdateOrganization.  # noqa: E501

        Mode for use of single sign on (SSO) for your organization.WARNING: It is dangerous to require SSO if you are not confident it is correctly configured. Once SSO is required, users may not log into Anyscale without using SSO. If you lock all users out of your organization by requiring incorrectly configured SSO, you may need to contact Anyscale for support.  # noqa: E501

        :return: The sso_mode of this UpdateOrganization.  # noqa: E501
        :rtype: SSOMode
        """
        return self._sso_mode

    @sso_mode.setter
    def sso_mode(self, sso_mode):
        """Sets the sso_mode of this UpdateOrganization.

        Mode for use of single sign on (SSO) for your organization.WARNING: It is dangerous to require SSO if you are not confident it is correctly configured. Once SSO is required, users may not log into Anyscale without using SSO. If you lock all users out of your organization by requiring incorrectly configured SSO, you may need to contact Anyscale for support.  # noqa: E501

        :param sso_mode: The sso_mode of this UpdateOrganization.  # noqa: E501
        :type: SSOMode
        """
        if self.local_vars_configuration.client_side_validation and sso_mode is None:  # noqa: E501
            raise ValueError("Invalid value for `sso_mode`, must not be `None`")  # noqa: E501

        self._sso_mode = sso_mode

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, UpdateOrganization):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, UpdateOrganization):
            return True

        return self.to_dict() != other.to_dict()



"""
This file holds all of the CLI commands for the "anyscale machine-pool" path.
"""

import click
import tabulate

from anyscale.controllers.machine_pool_controller import MachinePoolController


@click.group(
    "machine-pool", help="Commands to interact with machine pools in Anyscale.",
)
def machine_pool_cli() -> None:
    pass


@machine_pool_cli.command(name="create", help="Create a machine pool in Anyscale.")
@click.option(
    "--name",
    type=str,
    required=True,
    help="Provide a machine pool name (must be unique within an organization).",
)
def create_machine_pool(name: str) -> None:
    machine_pool_controller = MachinePoolController()
    output = machine_pool_controller.create_machine_pool(machine_pool_name=name)
    print(
        f"Machine pool {output.machine_pool.machine_pool_name} has been created successfully (ID {output.machine_pool.machine_pool_id})."
    )


@machine_pool_cli.command(name="delete", help="Delete a machine pool in Anyscale.")
@click.option("--name", type=str, required=True, help="Provide a machine pool name.")
def delete_machine_pool(name: str) -> None:
    machine_pool_controller = MachinePoolController()
    machine_pool_controller.delete_machine_pool(machine_pool_name=name)
    print(f"Deleted machine pool '{name}'.")


@machine_pool_cli.command(name="list", help="List machine pools in Anyscale.")
def list_machine_pools() -> None:
    machine_pool_controller = MachinePoolController()
    output = machine_pool_controller.list_machine_pools()
    table = []
    columns = [
        "MACHINE POOL",
        "ID",
        "Clouds",
    ]
    for mp in output.machine_pools:
        table.append(
            [mp.machine_pool_name, mp.machine_pool_id, "\n".join(mp.cloud_ids),]
        )
    print(tabulate.tabulate(table, tablefmt="plain", headers=columns, stralign="left"))


@machine_pool_cli.command(name="attach", help="Attach a machine pool to a cloud.")
@click.option("--name", type=str, required=True, help="Provide a machine pool name.")
@click.option("--cloud", type=str, required=True, help="Provide a cloud name.")
def attach_machine_pool_to_cloud(name: str, cloud: str) -> None:
    machine_pool_controller = MachinePoolController()
    machine_pool_controller.attach_machine_pool_to_cloud(
        machine_pool_name=name, cloud=cloud
    )
    print(f"Attached machine pool '{name}' to cloud '{cloud}'.")


@machine_pool_cli.command(name="detach", help="Detach a machine pool from a cloud.")
@click.option("--name", type=str, required=True, help="Provide a machine pool name.")
@click.option("--cloud", type=str, required=True, help="Provide a cloud name.")
def detach_machine_pool_from_cloud(name: str, cloud: str) -> None:
    machine_pool_controller = MachinePoolController()
    machine_pool_controller.detach_machine_pool_from_cloud(
        machine_pool_name=name, cloud=cloud
    )
    print(f"Detached machine pool '{name}' from cloud '{cloud}'.")



from unittest.mock import Mock

from anyscale.client.openapi_client.models import (
    ListMachinesResponse,
    MachineAllocationState,
    MachineConnectionState,
    MachineInfo,
)
from anyscale.controllers.machine_controller import MachineController


def test_list_machines(mock_auth_api_client: str) -> None:
    response = ListMachinesResponse(
        machines=[
            MachineInfo(
                machine_id="m-123",
                hostname="dummy",
                machine_shape="lambda-a100-80g",
                connection_state=MachineConnectionState.CONNECTED,
                allocation_state=MachineAllocationState.AVAILABLE,
            )
        ]
    )

    api_response = Mock()
    api_response.result = response

    machine_controller = MachineController()
    machine_controller.api_client.list_machines_api_v2_machines_get = Mock(
        return_value=api_response
    )
    output = machine_controller.list_machines(machine_pool_name="pool1")
    assert output == response



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CreateUser(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'str',
        'username': 'str',
        'email': 'str',
        'password': 'str',
        'invite_code': 'str',
        'organization_name': 'str',
        'invitation_id': 'str',
        'lastname': 'str',
        'magic_token': 'str',
        'title': 'str',
        'company': 'str',
        'organization_marketing_questions': 'OrganizationMarketingQuestions',
        'is_oa': 'bool',
        'captcha_token': 'str'
    }

    attribute_map = {
        'name': 'name',
        'username': 'username',
        'email': 'email',
        'password': 'password',
        'invite_code': 'invite_code',
        'organization_name': 'organization_name',
        'invitation_id': 'invitation_id',
        'lastname': 'lastname',
        'magic_token': 'magic_token',
        'title': 'title',
        'company': 'company',
        'organization_marketing_questions': 'organization_marketing_questions',
        'is_oa': 'is_oa',
        'captcha_token': 'captcha_token'
    }

    def __init__(self, name=None, username=None, email=None, password=None, invite_code=None, organization_name=None, invitation_id=None, lastname=None, magic_token=None, title=None, company=None, organization_marketing_questions=None, is_oa=False, captcha_token=None, local_vars_configuration=None):  # noqa: E501
        """CreateUser - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._username = None
        self._email = None
        self._password = None
        self._invite_code = None
        self._organization_name = None
        self._invitation_id = None
        self._lastname = None
        self._magic_token = None
        self._title = None
        self._company = None
        self._organization_marketing_questions = None
        self._is_oa = None
        self._captcha_token = None
        self.discriminator = None

        self.name = name
        self.username = username
        self.email = email
        if password is not None:
            self.password = password
        if invite_code is not None:
            self.invite_code = invite_code
        if organization_name is not None:
            self.organization_name = organization_name
        if invitation_id is not None:
            self.invitation_id = invitation_id
        if lastname is not None:
            self.lastname = lastname
        if magic_token is not None:
            self.magic_token = magic_token
        if title is not None:
            self.title = title
        if company is not None:
            self.company = company
        if organization_marketing_questions is not None:
            self.organization_marketing_questions = organization_marketing_questions
        if is_oa is not None:
            self.is_oa = is_oa
        if captcha_token is not None:
            self.captcha_token = captcha_token

    @property
    def name(self):
        """Gets the name of this CreateUser.  # noqa: E501


        :return: The name of this CreateUser.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this CreateUser.


        :param name: The name of this CreateUser.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def username(self):
        """Gets the username of this CreateUser.  # noqa: E501


        :return: The username of this CreateUser.  # noqa: E501
        :rtype: str
        """
        return self._username

    @username.setter
    def username(self, username):
        """Sets the username of this CreateUser.


        :param username: The username of this CreateUser.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and username is None:  # noqa: E501
            raise ValueError("Invalid value for `username`, must not be `None`")  # noqa: E501

        self._username = username

    @property
    def email(self):
        """Gets the email of this CreateUser.  # noqa: E501


        :return: The email of this CreateUser.  # noqa: E501
        :rtype: str
        """
        return self._email

    @email.setter
    def email(self, email):
        """Sets the email of this CreateUser.


        :param email: The email of this CreateUser.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and email is None:  # noqa: E501
            raise ValueError("Invalid value for `email`, must not be `None`")  # noqa: E501

        self._email = email

    @property
    def password(self):
        """Gets the password of this CreateUser.  # noqa: E501

        Password to use for logging in as the user. At least one authentication method should be used. Either password or magic link.  # noqa: E501

        :return: The password of this CreateUser.  # noqa: E501
        :rtype: str
        """
        return self._password

    @password.setter
    def password(self, password):
        """Sets the password of this CreateUser.

        Password to use for logging in as the user. At least one authentication method should be used. Either password or magic link.  # noqa: E501

        :param password: The password of this CreateUser.  # noqa: E501
        :type: str
        """

        self._password = password

    @property
    def invite_code(self):
        """Gets the invite_code of this CreateUser.  # noqa: E501


        :return: The invite_code of this CreateUser.  # noqa: E501
        :rtype: str
        """
        return self._invite_code

    @invite_code.setter
    def invite_code(self, invite_code):
        """Sets the invite_code of this CreateUser.


        :param invite_code: The invite_code of this CreateUser.  # noqa: E501
        :type: str
        """

        self._invite_code = invite_code

    @property
    def organization_name(self):
        """Gets the organization_name of this CreateUser.  # noqa: E501


        :return: The organization_name of this CreateUser.  # noqa: E501
        :rtype: str
        """
        return self._organization_name

    @organization_name.setter
    def organization_name(self, organization_name):
        """Sets the organization_name of this CreateUser.


        :param organization_name: The organization_name of this CreateUser.  # noqa: E501
        :type: str
        """

        self._organization_name = organization_name

    @property
    def invitation_id(self):
        """Gets the invitation_id of this CreateUser.  # noqa: E501


        :return: The invitation_id of this CreateUser.  # noqa: E501
        :rtype: str
        """
        return self._invitation_id

    @invitation_id.setter
    def invitation_id(self, invitation_id):
        """Sets the invitation_id of this CreateUser.


        :param invitation_id: The invitation_id of this CreateUser.  # noqa: E501
        :type: str
        """

        self._invitation_id = invitation_id

    @property
    def lastname(self):
        """Gets the lastname of this CreateUser.  # noqa: E501


        :return: The lastname of this CreateUser.  # noqa: E501
        :rtype: str
        """
        return self._lastname

    @lastname.setter
    def lastname(self, lastname):
        """Sets the lastname of this CreateUser.


        :param lastname: The lastname of this CreateUser.  # noqa: E501
        :type: str
        """

        self._lastname = lastname

    @property
    def magic_token(self):
        """Gets the magic_token of this CreateUser.  # noqa: E501

        Magic link token to authenticate ownership of an email address. At least one authentication method should be used. Either password or magic link.  # noqa: E501

        :return: The magic_token of this CreateUser.  # noqa: E501
        :rtype: str
        """
        return self._magic_token

    @magic_token.setter
    def magic_token(self, magic_token):
        """Sets the magic_token of this CreateUser.

        Magic link token to authenticate ownership of an email address. At least one authentication method should be used. Either password or magic link.  # noqa: E501

        :param magic_token: The magic_token of this CreateUser.  # noqa: E501
        :type: str
        """

        self._magic_token = magic_token

    @property
    def title(self):
        """Gets the title of this CreateUser.  # noqa: E501


        :return: The title of this CreateUser.  # noqa: E501
        :rtype: str
        """
        return self._title

    @title.setter
    def title(self, title):
        """Sets the title of this CreateUser.


        :param title: The title of this CreateUser.  # noqa: E501
        :type: str
        """

        self._title = title

    @property
    def company(self):
        """Gets the company of this CreateUser.  # noqa: E501


        :return: The company of this CreateUser.  # noqa: E501
        :rtype: str
        """
        return self._company

    @company.setter
    def company(self, company):
        """Sets the company of this CreateUser.


        :param company: The company of this CreateUser.  # noqa: E501
        :type: str
        """

        self._company = company

    @property
    def organization_marketing_questions(self):
        """Gets the organization_marketing_questions of this CreateUser.  # noqa: E501


        :return: The organization_marketing_questions of this CreateUser.  # noqa: E501
        :rtype: OrganizationMarketingQuestions
        """
        return self._organization_marketing_questions

    @organization_marketing_questions.setter
    def organization_marketing_questions(self, organization_marketing_questions):
        """Sets the organization_marketing_questions of this CreateUser.


        :param organization_marketing_questions: The organization_marketing_questions of this CreateUser.  # noqa: E501
        :type: OrganizationMarketingQuestions
        """

        self._organization_marketing_questions = organization_marketing_questions

    @property
    def is_oa(self):
        """Gets the is_oa of this CreateUser.  # noqa: E501


        :return: The is_oa of this CreateUser.  # noqa: E501
        :rtype: bool
        """
        return self._is_oa

    @is_oa.setter
    def is_oa(self, is_oa):
        """Sets the is_oa of this CreateUser.


        :param is_oa: The is_oa of this CreateUser.  # noqa: E501
        :type: bool
        """

        self._is_oa = is_oa

    @property
    def captcha_token(self):
        """Gets the captcha_token of this CreateUser.  # noqa: E501


        :return: The captcha_token of this CreateUser.  # noqa: E501
        :rtype: str
        """
        return self._captcha_token

    @captcha_token.setter
    def captcha_token(self, captcha_token):
        """Sets the captcha_token of this CreateUser.


        :param captcha_token: The captcha_token of this CreateUser.  # noqa: E501
        :type: str
        """

        self._captcha_token = captcha_token

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreateUser):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CreateUser):
            return True

        return self.to_dict() != other.to_dict()



from unittest.mock import Mock

from anyscale.utils.connect_helpers import search_entities


def test_search_entities():
    # 1. Test all records returned
    continued_response_of_size_two = Mock(
        results=[Mock(), Mock()], metadata=Mock(next_paging_token="token",),
    )
    mock_search_api = Mock(
        side_effect=[
            continued_response_of_size_two,
            Mock(results=[Mock()], metadata=Mock(next_paging_token=None,),),
        ]
    )
    assert len(search_entities(mock_search_api, query_obj=Mock())) == 3
    # 2. Test `max_to_return` reached
    mock_search_api = Mock(return_value=continued_response_of_size_two)
    assert len(search_entities(mock_search_api, query_obj=Mock(), max_to_return=1)) == 1



import asyncio
from typing import List, Optional, Tuple

from anyscale.api_utils.logs_util import (
    _download_logs_concurrently,
    _remove_ansi_escape_sequences,
)
from anyscale.cli_logger import LogsLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.models.base_job_status import BaseJobStatus
from anyscale.controllers.logs_controller import DEFAULT_PARALLELISM
from anyscale.sdk.anyscale_client.models.log_download_result import LogDownloadResult


_JOB_RUN_TERMINAL_STATES = {
    BaseJobStatus.SUCCEEDED,
    BaseJobStatus.STOPPED,
    BaseJobStatus.COMPLETED,
    BaseJobStatus.FAILED,
    BaseJobStatus.UNKNOWN,
}


async def _get_job_logs_from_storage_bucket_streaming(
    api: DefaultApi,
    log: LogsLogger,
    *,
    next_page_token: Optional[str] = None,
    job_run_id: Optional[str] = None,
    parallelism: int = DEFAULT_PARALLELISM,
    remove_escape_chars: bool = True,
    cluster_journal_events_start_line: int = 0,
    no_cluster_journal_events: bool = False,
) -> Tuple[str, Optional[str], int, bool]:
    """Retrieves logs directly from the storage bucket.
    Also fetches cluster journal logs and print them to stderr inline with job logs.

    Args:
        parallelism (int, optional): Defaults to 10 (AWS S3's default `max_concurrent_requests` value)
            There is no documented upper limit. However, please set a reasonable value.
        remove_escape_chars (bool, optional): Removes ANSI escape sequences from the logs, which are
            commonly used to color terminal output. Defaults to True.

    Returns (logs, next_page_token, cluster_journal_events_new_start_line, is_job_run_terminated):
        Returns the logs output and the next page token for fetching additional logs.
        cluster_journal_events_new_start_line is the new start line for fetching cluster journal events.
        is_job_run_terminated is True if the job run is terminated, False otherwise.

    Raises:
        NoJobRunError: If a job run hasn't been created yet.
    """
    # First fetch both the job info and the log chunks in parallel

    job_run_thread = api.get_decorated_job_api_v2_decorated_jobs_job_id_get(
        job_id=job_run_id, async_req=True
    )

    log_download_result_thread = api.get_job_logs_download_v2_api_v2_logs_job_logs_download_v2_job_id_get(
        job_id=job_run_id, next_page_token=next_page_token, async_req=True
    )

    # Once job info is fetched, fetch the cluster journal events
    job_run = job_run_thread.get().result
    cluster_id = job_run.cluster.id

    if not no_cluster_journal_events:
        cluster_journal_events_thread = api.get_startup_logs_api_v2_sessions_session_id_startup_logs_get(
            cluster_id,
            start_line=cluster_journal_events_start_line,
            end_line=10000000,
            async_req=True,
        )

    # Wait for the log chunks API and start downloading the chunks in parallel
    log_download_result: LogDownloadResult = log_download_result_thread.get().result
    all_log_chunk_urls: List[str] = [chunk.chunk_url for chunk in log_download_result.log_chunks]  # type: ignore
    logs_task = asyncio.create_task(
        _download_logs_concurrently(
            all_log_chunk_urls,
            parallelism,
            bearer_token=log_download_result.bearer_token,
        )
    )

    # Print out the cluster journal events
    if not no_cluster_journal_events:
        cluster_journal_events = cluster_journal_events_thread.get().result
        lines = cluster_journal_events.lines.splitlines()
        new_cluster_journal_events_start_line = (
            cluster_journal_events.start_line + cluster_journal_events.num_lines
        )
        if cluster_journal_events_start_line == 0:
            # When fetching the inital cluster journal events, we will just show the latest 10 events
            # to avoid flooding the output with too many events. In particular, after a job is
            # already running.
            lines = lines[-10:]
        for line in lines:
            log.info(line)
    else:
        new_cluster_journal_events_start_line = cluster_journal_events_start_line

    # Finally wait for all the log chunks to download and return all the logs
    logs = await logs_task
    if remove_escape_chars:
        logs = _remove_ansi_escape_sequences(logs)
    new_next_page_token = (
        log_download_result.next_page_token
        if len(log_download_result.log_chunks) > 0
        else next_page_token
    )
    return (
        logs,
        new_next_page_token,
        new_cluster_journal_events_start_line,
        job_run.status in _JOB_RUN_TERMINAL_STATES,
    )



from dataclasses import dataclass
import os
import re
from typing import Optional

import pytest

from anyscale.compute_config.models import ComputeConfig, HeadNodeConfig
from anyscale.service.models import (
    RayGCSExternalStorageConfig,
    ServiceConfig,
    ServiceState,
    ServiceStatus,
    ServiceVersionState,
    ServiceVersionStatus,
)


@dataclass
class ServiceConfigFile:
    name: str
    expected_config: Optional[ServiceConfig] = None
    expected_error: Optional[str] = None

    def get_path(self) -> str:
        return os.path.join(
            os.path.dirname(__file__), "test_files/service_config_files", self.name
        )


TEST_CONFIG_FILES = [
    ServiceConfigFile(
        "minimal.yaml",
        expected_config=ServiceConfig(applications=[{"import_path": "main:app"}]),
    ),
    ServiceConfigFile(
        "full.yaml",
        expected_config=ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-name-from-file",
            image_uri="docker.io/library/test:latest",
            compute_config=ComputeConfig(
                cloud="test-cloud", head_node=HeadNodeConfig(instance_type="m5.2xlarge")
            ),
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            query_auth_token_enabled=False,
            http_options={"request_timeout_s": 10.0},
            grpc_options={"grpc_servicer_functions": ["hello.world"]},
            logging_config={"log_level": "DEBUG"},
        ),
    ),
    ServiceConfigFile(
        "points_to_requirements_file.yaml",
        expected_config=ServiceConfig(
            applications=[{"import_path": "main:app"}],
            requirements="some_requirements_file.txt",
        ),
    ),
    ServiceConfigFile(
        "unrecognized_option.yaml",
        expected_error=re.escape(
            "__init__() got an unexpected keyword argument 'bad_option'"
        ),
    ),
    ServiceConfigFile(
        "multiple_applications.yaml",
        expected_config=ServiceConfig(
            applications=[
                {
                    "name": "app1",
                    "import_path": "main:app1",
                    "runtime_env": {"env_vars": {"abc": "def"}},
                },
                {
                    "name": "app2",
                    "import_path": "main:app",
                    "args": {"abc": "def", "nested": {"key": "val"}},
                },
            ],
        ),
    ),
]


class TestServiceConfig:
    def test_empty_applications(self):
        with pytest.raises(ValueError, match="'applications' cannot be empty."):
            ServiceConfig()

        with pytest.raises(ValueError, match="'applications' cannot be empty."):
            ServiceConfig(applications=[])

    def test_bad_applications_type(self):
        with pytest.raises(TypeError, match="'applications' must be a list."):
            ServiceConfig(applications={})

    def test_invalid_import_paths(self):
        with pytest.raises(TypeError, match="'import_path' must be a string"):
            ServiceConfig(applications=[{"import_path": 1}])

        with pytest.raises(ValueError, match="'import_path' must be"):
            ServiceConfig(applications=[{"import_path": "hello"}])

        with pytest.raises(ValueError, match="'import_path' must be"):
            ServiceConfig(applications=[{"import_path": "hello."}])

        with pytest.raises(ValueError, match="'import_path' must be"):
            ServiceConfig(applications=[{"import_path": "hello:"}])

        with pytest.raises(ValueError, match="'import_path' must be"):
            ServiceConfig(applications=[{"import_path": ":hello"}])

        with pytest.raises(ValueError, match="'import_path' must be"):
            ServiceConfig(applications=[{"import_path": ".hello"}])

    def test_query_auth_token_enabled(self):
        config = ServiceConfig(applications=[{"import_path": "main:app"}],)
        assert config.query_auth_token_enabled is True

        config = ServiceConfig(applications=[{"import_path": "main:app"}],)
        assert config.query_auth_token_enabled is True

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}], query_auth_token_enabled=False,
        )
        assert config.query_auth_token_enabled is False
        assert config.options().query_auth_token_enabled is False
        assert (
            config.options(query_auth_token_enabled=True).query_auth_token_enabled
            is True
        )

        with pytest.raises(
            TypeError, match="'query_auth_token_enabled' must be a boolean."
        ):
            ServiceConfig(
                applications=[{"import_path": "main:app"}],
                query_auth_token_enabled="foobar",
            )

    def test_http_options(self):
        config = ServiceConfig(applications=[{"import_path": "main:app"}],)
        assert config.http_options is None

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            http_options={"request_timeout_s": 10.0},
        )
        assert config.http_options == {"request_timeout_s": 10.0}

        with pytest.raises(TypeError, match="'http_options' must be a dict."):
            ServiceConfig(
                applications=[{"import_path": "main:app"}],
                http_options=["request_timeout_s"],
            )

        with pytest.raises(
            ValueError,
            match="The following provided 'http_options' are not permitted in Anyscale: {'host'}.",
        ):
            ServiceConfig(
                applications=[{"import_path": "main:app"}],
                http_options={"request_timeout_s": 10.0, "host": "0.0.0.0"},
            )

    def test_grpc_options(self):
        config = ServiceConfig(applications=[{"import_path": "main:app"}],)
        assert config.grpc_options is None

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            grpc_options={"grpc_servicer_functions": ["hello.world"]},
        )
        assert config.grpc_options == {"grpc_servicer_functions": ["hello.world"]}

        with pytest.raises(TypeError, match="'grpc_options' must be a dict."):
            ServiceConfig(
                applications=[{"import_path": "main:app"}],
                grpc_options=["grpc_servicer_functions"],
            )

        with pytest.raises(
            ValueError,
            match="The following provided 'http_options' are not permitted in Anyscale: {'port'}.",
        ):
            ServiceConfig(
                applications=[{"import_path": "main:app"}],
                http_options={"grpc_servicer_functions": ["hello.world"], "port": 9001},
            )

    def test_logging_config(self):
        config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert config.logging_config is None

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            logging_config={"log_level": "DEBUG"},
        )
        assert config.logging_config == {"log_level": "DEBUG"}

        with pytest.raises(TypeError, match="'logging_config' must be a dict."):
            ServiceConfig(
                applications=[{"import_path": "main:app"}],
                logging_config=["log_level"],
            )

    def test_ray_gcs_external_storage_config(self):
        config = ServiceConfig(applications=[{"import_path": "main:app"}])
        assert config.ray_gcs_external_storage_config is None

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            ray_gcs_external_storage_config=RayGCSExternalStorageConfig(
                address="host:1234"
            ),
        )
        assert config.ray_gcs_external_storage_config == RayGCSExternalStorageConfig(
            address="host:1234"
        )

        with pytest.raises(
            TypeError,
            match="'ray_gcs_external_storage_config' must be a RayGCSExternalStorageConfig.",
        ):
            ServiceConfig(
                applications=[{"import_path": "main:app"}],
                ray_gcs_external_storage_config=["address"],
            )

    def test_options(self):
        config = ServiceConfig(applications=[{"import_path": "main:app"}])

        options = {
            "name": "test-name",
            "image_uri": "docker.io/libaray/test-image:latest",
            "compute_config": "test-compute-config",
            "requirements": ["pip-install-test"],
            "working_dir": ".",
            "excludes": ["some-path"],
            "query_auth_token_enabled": False,
            "http_options": {"request_timeout_s": 10.0},
            "grpc_options": {"grpc_servicer_functions": ["hello.world"]},
            "logging_config": {"log_level": "DEBUG"},
            "ray_gcs_external_storage_config": RayGCSExternalStorageConfig(
                address="host:1234"
            ),
            "env_vars": {"k1": "v1"},
        }

        # Test setting fields one at a time.
        for option, val in options.items():
            assert config.options(**{option: val}) == ServiceConfig(
                applications=[{"import_path": "main:app"}], **{option: val}
            )

            assert config.options(**{option: val}) == ServiceConfig(
                applications=[{"import_path": "main:app"}], **{option: val}
            )

        # Test setting fields all at once.
        assert config.options(**options) == ServiceConfig(
            applications=[{"import_path": "main:app"}], **options
        )

    @pytest.mark.parametrize("config_file", TEST_CONFIG_FILES)
    def test_from_config_file(self, config_file: ServiceConfigFile):
        if config_file.expected_error is not None:
            with pytest.raises(Exception, match=config_file.expected_error):
                ServiceConfig.from_yaml(config_file.get_path())

            return

        assert config_file.expected_config == ServiceConfig.from_yaml(
            config_file.get_path()
        )


class TestServiceStatus:
    @pytest.mark.parametrize(
        "canary_version",
        [
            None,
            ServiceVersionStatus(
                state=ServiceVersionState.STARTING,
                id="test-canary-version",
                weight=0,
                config=ServiceConfig(
                    applications=[
                        {"import_path": "main:app", "args": {"version": "canary"}}
                    ],
                    image_uri="docker.io/library/test-image:latest",
                    compute_config="test-compute-config",
                ),
            ),
        ],
    )
    @pytest.mark.parametrize(
        "primary_version",
        [
            None,
            ServiceVersionStatus(
                state=ServiceVersionState.RUNNING,
                id="test-primary-version",
                weight=100,
                config=ServiceConfig(
                    applications=[
                        {"import_path": "main:app", "args": {"version": "primary"}}
                    ],
                    image_uri="docker.io/library/test-image:latest",
                    compute_config="test-compute-config",
                ),
            ),
        ],
    )
    @pytest.mark.parametrize("query_auth_token", [None, "test_abc123"])
    @pytest.mark.parametrize("state", list(ServiceState))  # type: ignore
    def test_to_dict_and_back(
        self,
        state: ServiceState,
        query_auth_token: Optional[str],
        primary_version: Optional[ServiceVersionStatus],
        canary_version: Optional[ServiceVersionStatus],
    ):
        """Test that all fields can be serialized to and from dictionaries."""
        status = ServiceStatus(
            id="test-service-id",
            name="test-service-name",
            state=state,
            query_url="http://test.com/",
            query_auth_token=query_auth_token,
            primary_version=primary_version,
            canary_version=canary_version,
        )

        assert ServiceStatus.from_dict(status.to_dict()) == status

    @pytest.mark.parametrize("state", list(ServiceVersionState))  # type: ignore
    def test_version_states(self, state: ServiceVersionState):
        assert (
            ServiceVersionStatus(
                id="test-version",
                state=state,
                weight=100,
                config=ServiceConfig(
                    applications=[
                        {"import_path": "main:app", "args": {"version": "primary"}}
                    ],
                ),
            ).state
            == state
        )

    def test_unknown_states(self):
        with pytest.raises(
            ValueError, match="'SOME_FAKE_NEWS_STATE' is not a valid ServiceState"
        ):
            ServiceStatus(
                id="test-service-id",
                name="test-service-name",
                state="SOME_FAKE_NEWS_STATE",
                query_url="http://test.com/",
                query_auth_token=None,
                primary_version=None,
                canary_version=None,
            )

        with pytest.raises(
            ValueError,
            match="'SOME_FAKE_NEWS_STATE' is not a valid ServiceVersionState",
        ):
            ServiceVersionStatus(
                id="test-primary-version",
                state="SOME_FAKE_NEWS_STATE",
                weight=100,
                config=ServiceConfig(
                    applications=[
                        {"import_path": "main:app", "args": {"version": "primary"}}
                    ],
                    image_uri="docker.io/library/test-image:latest",
                    compute_config="test-compute-config",
                ),
            ),



import click
import pytest

from anyscale.utils.imports.all import try_import_ray
from anyscale.utils.imports.gcp import try_import_gcp_secretmanager


def test_try_import_ray():
    # This test verifies  that try_import_ray() runs successfully
    # if Ray is installed and returns the same version of Ray as
    # the local import statement. If Ray is not installed, try_import_ray()
    # should raise an error.

    # TODO(nikita): This test should be run in an environment that has Ray
    # installed, and one that doesn't have Ray installed. The
    # `import ray` statement inside the method cannot be mocked.
    try:
        import ray

        assert try_import_ray() == ray
    except ImportError:
        with pytest.raises(click.ClickException):
            try_import_ray()


def test_try_import_gcp_secretmanager():
    try:
        from google.cloud import secretmanager

        assert try_import_gcp_secretmanager() == secretmanager
    except ImportError:
        with pytest.raises(click.ClickException):
            try_import_gcp_secretmanager()
